#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import pytest

import octobot_trading.constants as constants
import octobot_trading.errors as errors
import octobot_trading.personal_data as trading_personal_data

import octobot_commons.constants as commons_constants

from tests.exchanges import backtesting_trader, backtesting_config, backtesting_exchange_manager, fake_backtesting


def test_try_convert_currency_value_using_multiple_pairs(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    value_converter = portfolio_manager.portfolio_value_holder.value_converter

    # try_convert_currency_value_using_multiple_pairs uses last_prices_by_trading_pair
    # try without last_prices_by_trading_pair
    assert value_converter.try_convert_currency_value_using_multiple_pairs("ETH", "BTC", constants.ONE, []) \
           is None
    assert value_converter.try_convert_currency_value_using_multiple_pairs("BTC", "BTC", constants.ONE, []) \
           is None
    assert value_converter.try_convert_currency_value_using_multiple_pairs("USDT", "BTC", constants.ONE, []) \
           is None
    assert value_converter.try_convert_currency_value_using_multiple_pairs("DOT", "BTC", constants.ONE, []) \
           is None
    assert value_converter.try_convert_currency_value_using_multiple_pairs("ADA", "BTC", constants.ONE, []) \
           is None

    # 1 ETH = 0.5 BTC
    value_converter.last_prices_by_trading_pair["ETH/BTC"] = decimal.Decimal("0.5")
    # 1 DOT = 0.2 ETH
    value_converter.update_last_price("DOT/ETH", decimal.Decimal("0.2"))
    # therefore 1 DOT = 0.2 ETH = 0.5 * 0.2 BTC = 0.1 BTC
    assert value_converter.try_convert_currency_value_using_multiple_pairs("DOT", "BTC", decimal.Decimal("2"), []) \
        == decimal.Decimal("0.2")

    # with reversed pair in bridge
    value_converter.last_prices_by_trading_pair["BTC/USDT"] = decimal.Decimal("10000")
    value_converter.last_prices_by_trading_pair["ADA/USDT"] = decimal.Decimal("2")
    # 1 ADA = 2 USDT == 2/10000 BTC
    assert value_converter.try_convert_currency_value_using_multiple_pairs("ADA", "BTC", decimal.Decimal(1), []) \
        == decimal.Decimal(2) / decimal.Decimal(10000)
    # now using bridges cache
    assert value_converter.try_convert_currency_value_using_multiple_pairs("ADA", "BTC", decimal.Decimal(2), []) \
        == decimal.Decimal(4) / decimal.Decimal(10000)
    # bridge is saved
    assert value_converter.get_saved_price_conversion_bridge("ADA", "BTC") == [
        ("ADA", "USDT"), ("USDT", "BTC")
    ]
    assert value_converter.convert_currency_value_from_saved_price_bridges(
        "ADA", "BTC", decimal.Decimal(10)
    ) == decimal.Decimal(20) / decimal.Decimal(10000)
    # without enough data
    value_converter.last_prices_by_trading_pair["CRO/PLOP"] = decimal.Decimal("2")
    assert value_converter.try_convert_currency_value_using_multiple_pairs("CRO", "BTC", constants.ONE, []) \
           is None
    # second time to make sure bridge cache is not creating issues
    assert value_converter.try_convert_currency_value_using_multiple_pairs("CRO", "BTC", constants.ONE, []) \
           is None
    with pytest.raises(KeyError):
        # no bridge saved as value is not available
        value_converter.get_saved_price_conversion_bridge("CRO", "BTC")
    with pytest.raises(errors.MissingPriceDataError):
        value_converter.convert_currency_value_from_saved_price_bridges(
            "CRO", "BTC", constants.ONE
        )
    with pytest.raises(errors.MissingPriceDataError):
        value_converter.convert_currency_value_from_saved_price_bridges(
            "PLOP", "BTC", constants.ONE
        )

    exchange_manager.exchange_config.traded_symbol_pairs.append("NANO/BTC")
    # part 1 of bridge data that has not been update but are in exchange config therefore will be available
    with pytest.raises(errors.PendingPriceDataError):
        value_converter.try_convert_currency_value_using_multiple_pairs("NANO", "BTC", constants.ONE, [])

    assert value_converter.try_convert_currency_value_using_multiple_pairs("XRP", "BTC", constants.ONE, []) \
           is None
    # provide first part of the bridge
    exchange_manager.exchange_config.traded_symbol_pairs.append("XRP/USDT")
    assert value_converter.is_missing_price_bridge("XRP", "BTC")
    assert value_converter.try_convert_currency_value_using_multiple_pairs("XRP", "BTC", constants.ONE, []) \
           is None
    value_converter.reset_missing_price_bridges()
    value_converter.update_last_price("DOT/ETH", decimal.Decimal("0.2"))
    with pytest.raises(errors.PendingPriceDataError):
        value_converter.try_convert_currency_value_using_multiple_pairs("XRP", "BTC", constants.ONE, [])


def test_try_convert_currency_value_using_multiple_pairs_with_nested_bridges(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    value_converter = portfolio_manager.portfolio_value_holder.value_converter

    assert value_converter.try_convert_currency_value_using_multiple_pairs("XRP", "BTC", constants.ONE, []) \
           is None
    # provide first part 1 of the bridge
    value_converter.last_prices_by_trading_pair["ADA/USDT"] = decimal.Decimal("2")
    exchange_manager.exchange_config.traded_symbol_pairs.append("BTC/USDT")
    with pytest.raises(errors.PendingPriceDataError):
        # missing part 2 price
        value_converter.try_convert_currency_value_using_multiple_pairs("ADA", "BTC", constants.ONE, [])
    value_converter.last_prices_by_trading_pair["BTC/USDT"] = decimal.Decimal("100")
    assert value_converter.try_convert_currency_value_using_multiple_pairs("ADA", "BTC", constants.ONE, []) == \
           decimal.Decimal(2) / decimal.Decimal(100)
    assert value_converter.get_saved_price_conversion_bridge("ADA", "BTC") == [
        ("ADA", "USDT"), ("USDT", "BTC")
    ]

    # need to go through BTC to get ETH price
    value_converter.last_prices_by_trading_pair["ETH/BTC"] = decimal.Decimal("0.1")
    assert value_converter.try_convert_currency_value_using_multiple_pairs("ADA", "ETH", constants.ONE, []) == \
           decimal.Decimal(2) / decimal.Decimal(100) / decimal.Decimal("0.1")
    assert value_converter.get_saved_price_conversion_bridge("ADA", "ETH") == [
        ("ADA", "USDT"), ("USDT", "BTC"), ("BTC", "ETH")
    ]
    assert value_converter.convert_currency_value_from_saved_price_bridges("ADA", "ETH", decimal.Decimal("1.2")) \
        == decimal.Decimal("1.2") * decimal.Decimal("2") / decimal.Decimal("100") / decimal.Decimal("0.1")
    # also saved intermediary bridges
    assert value_converter.get_saved_price_conversion_bridge("USDT", "ETH") == [
        ("USDT", "BTC"), ("BTC", "ETH")
    ]
    assert value_converter.convert_currency_value_from_saved_price_bridges("USDT", "ETH", constants.ONE) \
        == constants.ONE / decimal.Decimal("100") / decimal.Decimal("0.1")

    # need to go through BTC and ETH to get ETH XRP price
    value_converter.last_prices_by_trading_pair["XRP/ETH"] = decimal.Decimal("0.0000001")
    assert value_converter.try_convert_currency_value_using_multiple_pairs("ADA", "XRP", constants.ONE, []) == \
           decimal.Decimal(2) / decimal.Decimal(100) / decimal.Decimal("0.1") / decimal.Decimal("0.0000001")
    assert value_converter.get_saved_price_conversion_bridge("ADA", "XRP") == [
        ("ADA", "USDT"), ("USDT", "BTC"), ("BTC", "ETH"), ("ETH", "XRP")
    ]
    # now use bridges cache
    assert value_converter.try_convert_currency_value_using_multiple_pairs("ADA", "XRP", constants.ONE, []) == \
           decimal.Decimal(2) / decimal.Decimal(100) / decimal.Decimal("0.1") / decimal.Decimal("0.0000001")
    assert value_converter.convert_currency_value_from_saved_price_bridges(
        "ADA", "XRP", decimal.Decimal("0.0001")
    ) == decimal.Decimal("0.0001") * decimal.Decimal(2) / decimal.Decimal(100) / \
        decimal.Decimal("0.1") / decimal.Decimal("0.0000001")


def test_get_usd_like_value(backtesting_trader):
    config, exchange_manager, trader = backtesting_trader
    portfolio_manager = exchange_manager.exchange_personal_data.portfolio_manager
    value_converter = portfolio_manager.portfolio_value_holder.value_converter
    # no last_prices_by_trading_pair
    assert value_converter.get_usd_like_value(commons_constants.USD_LIKE_COINS[0], decimal.Decimal("11")) \
           == decimal.Decimal("11")
    assert value_converter.get_usd_like_value(commons_constants.USD_LIKE_COINS[-1], decimal.Decimal("11")) \
           == decimal.Decimal("11")
    with pytest.raises(errors.MissingPriceDataError):
        value_converter.get_usd_like_value("BTC", decimal.Decimal("11"))

    value_converter.update_last_price("BTC/USDC", decimal.Decimal("30000"))
    assert value_converter.get_usd_like_value("BTC", decimal.Decimal("11")) \
           == decimal.Decimal("11") * decimal.Decimal("30000")

    with pytest.raises(errors.MissingPriceDataError):
        value_converter.get_usd_like_value("ETH", decimal.Decimal("11"))


def test_can_convert_symbol_to_usd_like():
    assert trading_personal_data.ValueConverter.can_convert_symbol_to_usd_like("BTC/USDT") is True
    assert trading_personal_data.ValueConverter.can_convert_symbol_to_usd_like(
        f"{commons_constants.USD_LIKE_COINS[4]}/BTC"
    ) is True
    assert trading_personal_data.ValueConverter.can_convert_symbol_to_usd_like("BTC/ETH") is False



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
import asyncio

import octobot_trading.personal_data as personal_data


@pytest.fixture
def mock_callback():
    async def callback(*args):
        callback.called = True
        callback.args = args

    callback.called = False
    callback.args = None
    return callback

@pytest.fixture
def trigger(mock_callback):
    return personal_data.BaseTrigger(mock_callback, ("arg1", "arg2"))

def test_init(trigger, mock_callback):
    assert trigger.on_trigger_callback == mock_callback
    assert trigger.on_trigger_callback_args == ("arg1", "arg2")
    assert trigger._trigger_event is None
    assert trigger._trigger_task is None

def test_triggered_when_not_initialized(trigger):
    assert not trigger.triggered()

def test_is_pending_when_not_initialized(trigger):
    assert not trigger.is_pending()

def test_str_representation(trigger):
    expected = f"BaseTrigger({trigger.on_trigger_callback.__name__})"
    assert str(trigger) == expected
    trigger.clear()
    assert str(trigger) == f"BaseTrigger({None})"

@pytest.mark.asyncio
async def test_clear_cancels_pending_task(trigger):
    # Create a mock event and task
    trigger._trigger_event = asyncio.Event()
    trigger._trigger_task = asyncio.create_task(asyncio.sleep(1))

    # Clear should cancel the task and clear attributes
    assert trigger.on_trigger_callback is not None
    assert trigger.on_trigger_callback_args is not None
    trigger.clear()
    assert trigger._trigger_task is None
    assert trigger.on_trigger_callback is None
    assert trigger.on_trigger_callback_args is None

@pytest.mark.asyncio
async def test_call_callback(trigger, mock_callback):
    await trigger.call_callback()
    assert mock_callback.called
    assert mock_callback.args == ("arg1", "arg2")

def test_update_from_other_trigger(trigger):
    with pytest.raises(NotImplementedError):
        trigger.update_from_other_trigger(None)

@pytest.mark.asyncio
async def test_create_watcher_raises_not_implemented(trigger):
    with pytest.raises(NotImplementedError):
        await trigger.create_watcher()

# Test for concrete implementation requirements
def test_create_trigger_event_not_implemented(trigger):
    with pytest.raises(NotImplementedError):
        trigger._create_trigger_event()

@pytest.mark.asyncio
async def test_wait_for_trigger_set(trigger, mock_callback):
    trigger._trigger_event = asyncio.Event()
    wait_task = asyncio.create_task(trigger._wait_for_trigger_set())
    trigger._trigger_event.set()

    # Wait for the callback to be called
    await wait_task
    assert mock_callback.called
    assert mock_callback.args == ("arg1", "arg2")



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import pytest
import mock

import octobot_commons.signals as commons_signals
import octobot_trading.enums as enums
import octobot_trading.constants as constants
import octobot_trading.signals as signals
import octobot_trading.signals.util as signals_util

from tests import event_loop
from tests.exchanges import simulated_trader, simulated_exchange_manager
from tests.personal_data.orders import buy_limit_order, sell_limit_order, buy_market_order

import octobot_trading.personal_data as personal_data


@pytest.mark.asyncio
async def test_create_order_signal_description(buy_limit_order, sell_limit_order, buy_market_order):
    buy_market_order.reduce_only = True
    buy_market_order.tag = "hello"
    buy_market_order.symbol = "BTC/ETH"
    buy_market_order.origin_quantity = decimal.Decimal("1.12")
    exchange_manager = buy_market_order.exchange_manager
    assert signals.create_order_signal_content(buy_market_order, enums.TradingSignalOrdersActions.CREATE,
                                               "strat", exchange_manager) == {
        enums.TradingSignalCommonsAttrs.ACTION.value: enums.TradingSignalOrdersActions.CREATE.value,
        enums.TradingSignalOrdersAttrs.SIDE.value: enums.TradeOrderSide.BUY.value,
        enums.TradingSignalOrdersAttrs.STRATEGY.value: "strat",
        enums.TradingSignalOrdersAttrs.SYMBOL.value: "BTC/ETH",
        enums.TradingSignalOrdersAttrs.EXCHANGE.value: "binanceus",
        enums.TradingSignalOrdersAttrs.EXCHANGE_TYPE.value: enums.ExchangeTypes.SPOT.value,
        enums.TradingSignalOrdersAttrs.TYPE.value: enums.TraderOrderType.BUY_MARKET.value,
        enums.TradingSignalOrdersAttrs.QUANTITY.value: float(buy_market_order.origin_quantity),
        enums.TradingSignalOrdersAttrs.TARGET_AMOUNT.value: None,
        enums.TradingSignalOrdersAttrs.TARGET_POSITION.value: None,
        enums.TradingSignalOrdersAttrs.TRIGGER_ABOVE.value: None,
        enums.TradingSignalOrdersAttrs.UPDATED_TARGET_AMOUNT.value: None,
        enums.TradingSignalOrdersAttrs.UPDATED_TARGET_POSITION.value: None,
        enums.TradingSignalOrdersAttrs.LIMIT_PRICE.value: 0.0,
        enums.TradingSignalOrdersAttrs.UPDATED_LIMIT_PRICE.value: 0.0,
        enums.TradingSignalOrdersAttrs.STOP_PRICE.value: 0.0,
        enums.TradingSignalOrdersAttrs.UPDATED_STOP_PRICE.value: 0.0,
        enums.TradingSignalOrdersAttrs.CURRENT_PRICE.value: 0.0,
        enums.TradingSignalOrdersAttrs.UPDATED_CURRENT_PRICE.value: 0.0,
        enums.TradingSignalOrdersAttrs.REDUCE_ONLY.value: True,
        enums.TradingSignalOrdersAttrs.POST_ONLY.value: False,
        enums.TradingSignalOrdersAttrs.GROUP_ID.value: None,
        enums.TradingSignalOrdersAttrs.GROUP_TYPE.value: None,
        enums.TradingSignalOrdersAttrs.ACTIVE_SWAP_STRATEGY_TYPE.value: None,
        enums.TradingSignalOrdersAttrs.ACTIVE_SWAP_STRATEGY_TIMEOUT.value: None,
        enums.TradingSignalOrdersAttrs.ACTIVE_SWAP_STRATEGY_TRIGGER_CONFIG.value: None,
        enums.TradingSignalOrdersAttrs.TRAILING_PROFILE_TYPE.value: None,
        enums.TradingSignalOrdersAttrs.TRAILING_PROFILE.value: None,
        enums.TradingSignalOrdersAttrs.IS_ACTIVE.value: True,
        enums.TradingSignalOrdersAttrs.ACTIVE_TRIGGER_PRICE.value: None,
        enums.TradingSignalOrdersAttrs.ACTIVE_TRIGGER_ABOVE.value: None,
        enums.TradingSignalOrdersAttrs.TAG.value: "hello",
        enums.TradingSignalOrdersAttrs.ORDER_ID.value: buy_market_order.order_id,
        enums.TradingSignalOrdersAttrs.BUNDLED_WITH.value: None,
        enums.TradingSignalOrdersAttrs.CHAINED_TO.value: None,
        enums.TradingSignalOrdersAttrs.ADDITIONAL_ORDERS.value: [],
        enums.TradingSignalOrdersAttrs.ASSOCIATED_ORDER_IDS.value: None,
        enums.TradingSignalOrdersAttrs.UPDATE_WITH_TRIGGERING_ORDER_FEES.value: False,
        enums.TradingSignalOrdersAttrs.CANCEL_POLICY_TYPE.value: None,
        enums.TradingSignalOrdersAttrs.CANCEL_POLICY_KWARGS.value: None,
    }


    buy_limit_order.reduce_only = True
    buy_limit_order.tag = "hello"
    buy_limit_order.symbol = "BTC/ETH"
    buy_limit_order.origin_price = decimal.Decimal("1.11")
    buy_limit_order.origin_quantity = decimal.Decimal("1.12")
    buy_limit_order.origin_stop_price = decimal.Decimal("1.13")
    buy_limit_order.is_active = False
    buy_limit_order.use_active_trigger(personal_data.create_order_price_trigger(buy_limit_order, decimal.Decimal("1.14"), False))
    buy_limit_order.cancel_policy = personal_data.create_cancel_policy(
        personal_data.ExpirationTimeOrderCancelPolicy.__name__,
        {
            "expiration_time": 999
        }
    )
    exchange_manager = buy_limit_order.exchange_manager
    assert signals.create_order_signal_content(buy_limit_order, enums.TradingSignalOrdersActions.CREATE,
                                               "strat", exchange_manager) == {
        enums.TradingSignalCommonsAttrs.ACTION.value: enums.TradingSignalOrdersActions.CREATE.value,
        enums.TradingSignalOrdersAttrs.SIDE.value: enums.TradeOrderSide.BUY.value,
        enums.TradingSignalOrdersAttrs.STRATEGY.value: "strat",
        enums.TradingSignalOrdersAttrs.SYMBOL.value: "BTC/ETH",
        enums.TradingSignalOrdersAttrs.EXCHANGE.value: "binanceus",
        enums.TradingSignalOrdersAttrs.EXCHANGE_TYPE.value: enums.ExchangeTypes.SPOT.value,
        enums.TradingSignalOrdersAttrs.TYPE.value: enums.TraderOrderType.BUY_LIMIT.value,
        enums.TradingSignalOrdersAttrs.QUANTITY.value: float(buy_limit_order.origin_quantity),
        enums.TradingSignalOrdersAttrs.TARGET_AMOUNT.value: None,
        enums.TradingSignalOrdersAttrs.TARGET_POSITION.value: None,
        enums.TradingSignalOrdersAttrs.TRIGGER_ABOVE.value: False,
        enums.TradingSignalOrdersAttrs.UPDATED_TARGET_AMOUNT.value: None,
        enums.TradingSignalOrdersAttrs.UPDATED_TARGET_POSITION.value: None,
        enums.TradingSignalOrdersAttrs.LIMIT_PRICE.value: float(buy_limit_order.origin_price),
        enums.TradingSignalOrdersAttrs.UPDATED_LIMIT_PRICE.value: 0.0,
        enums.TradingSignalOrdersAttrs.STOP_PRICE.value: float(buy_limit_order.origin_stop_price),
        enums.TradingSignalOrdersAttrs.UPDATED_STOP_PRICE.value: 0.0,
        enums.TradingSignalOrdersAttrs.CURRENT_PRICE.value: float(buy_limit_order.created_last_price),
        enums.TradingSignalOrdersAttrs.UPDATED_CURRENT_PRICE.value: 0.0,
        enums.TradingSignalOrdersAttrs.REDUCE_ONLY.value: True,
        enums.TradingSignalOrdersAttrs.POST_ONLY.value: False,
        enums.TradingSignalOrdersAttrs.GROUP_ID.value: None,
        enums.TradingSignalOrdersAttrs.GROUP_TYPE.value: None,
        enums.TradingSignalOrdersAttrs.ACTIVE_SWAP_STRATEGY_TYPE.value: None,
        enums.TradingSignalOrdersAttrs.ACTIVE_SWAP_STRATEGY_TIMEOUT.value: None,
        enums.TradingSignalOrdersAttrs.ACTIVE_SWAP_STRATEGY_TRIGGER_CONFIG.value: None,
        enums.TradingSignalOrdersAttrs.TRAILING_PROFILE_TYPE.value: None,
        enums.TradingSignalOrdersAttrs.TRAILING_PROFILE.value: None,
        enums.TradingSignalOrdersAttrs.IS_ACTIVE.value: False,
        enums.TradingSignalOrdersAttrs.ACTIVE_TRIGGER_PRICE.value: 1.14,
        enums.TradingSignalOrdersAttrs.ACTIVE_TRIGGER_ABOVE.value: False,
        enums.TradingSignalOrdersAttrs.TAG.value: "hello",
        enums.TradingSignalOrdersAttrs.ORDER_ID.value: buy_limit_order.order_id,
        enums.TradingSignalOrdersAttrs.BUNDLED_WITH.value: None,
        enums.TradingSignalOrdersAttrs.CHAINED_TO.value: None,
        enums.TradingSignalOrdersAttrs.ADDITIONAL_ORDERS.value: [],
        enums.TradingSignalOrdersAttrs.ASSOCIATED_ORDER_IDS.value: None,
        enums.TradingSignalOrdersAttrs.UPDATE_WITH_TRIGGERING_ORDER_FEES.value: False,
        enums.TradingSignalOrdersAttrs.CANCEL_POLICY_TYPE.value: personal_data.ExpirationTimeOrderCancelPolicy.__name__,
        enums.TradingSignalOrdersAttrs.CANCEL_POLICY_KWARGS.value: {
            "expiration_time": 999
        },
    }

    sell_limit_order.add_chained_order(buy_limit_order)
    sell_limit_order.symbol = "BTC/ETH"
    sell_limit_order.is_active = False
    sell_limit_order.use_active_trigger(personal_data.create_order_price_trigger(sell_limit_order, decimal.Decimal("2.14"), True))
    buy_limit_order.associate_to_entry("1")
    buy_limit_order.use_active_trigger(personal_data.create_order_price_trigger(buy_limit_order, decimal.Decimal("1.14"), True))
    buy_limit_order.cancel_policy = personal_data.create_cancel_policy(
        personal_data.ChainedOrderFillingPriceOrderCancelPolicy.__name__,
        {}
    )
    await buy_limit_order.set_as_chained_order(sell_limit_order, True, {}, True)
    assert signals.create_order_signal_content(
        buy_limit_order,
        enums.TradingSignalOrdersActions.CREATE,
        "strat",
        exchange_manager,
        target_amount="1%",
        target_position="2",
    ) == {
        enums.TradingSignalCommonsAttrs.ACTION.value: enums.TradingSignalOrdersActions.CREATE.value,
        enums.TradingSignalOrdersAttrs.SIDE.value: enums.TradeOrderSide.BUY.value,
        enums.TradingSignalOrdersAttrs.STRATEGY.value: "strat",
        enums.TradingSignalOrdersAttrs.SYMBOL.value: "BTC/ETH",
        enums.TradingSignalOrdersAttrs.EXCHANGE.value: "binanceus",
        enums.TradingSignalOrdersAttrs.EXCHANGE_TYPE.value: enums.ExchangeTypes.SPOT.value,
        enums.TradingSignalOrdersAttrs.TYPE.value: enums.TraderOrderType.BUY_LIMIT.value,
        enums.TradingSignalOrdersAttrs.QUANTITY.value: float(buy_limit_order.origin_quantity),
        enums.TradingSignalOrdersAttrs.TARGET_AMOUNT.value: "1%",
        enums.TradingSignalOrdersAttrs.TARGET_POSITION.value: "2",
        enums.TradingSignalOrdersAttrs.TRIGGER_ABOVE.value: False,
        enums.TradingSignalOrdersAttrs.UPDATED_TARGET_AMOUNT.value: None,
        enums.TradingSignalOrdersAttrs.UPDATED_TARGET_POSITION.value: None,
        enums.TradingSignalOrdersAttrs.LIMIT_PRICE.value: float(buy_limit_order.origin_price),
        enums.TradingSignalOrdersAttrs.UPDATED_LIMIT_PRICE.value: 0.0,
        enums.TradingSignalOrdersAttrs.STOP_PRICE.value: float(buy_limit_order.origin_stop_price),
        enums.TradingSignalOrdersAttrs.UPDATED_STOP_PRICE.value: 0.0,
        enums.TradingSignalOrdersAttrs.CURRENT_PRICE.value: float(buy_limit_order.created_last_price),
        enums.TradingSignalOrdersAttrs.UPDATED_CURRENT_PRICE.value: 0.0,
        enums.TradingSignalOrdersAttrs.REDUCE_ONLY.value: True,
        enums.TradingSignalOrdersAttrs.POST_ONLY.value: False,
        enums.TradingSignalOrdersAttrs.GROUP_ID.value: None,
        enums.TradingSignalOrdersAttrs.GROUP_TYPE.value: None,
        enums.TradingSignalOrdersAttrs.ACTIVE_SWAP_STRATEGY_TYPE.value: None,
        enums.TradingSignalOrdersAttrs.ACTIVE_SWAP_STRATEGY_TIMEOUT.value: None,
        enums.TradingSignalOrdersAttrs.ACTIVE_SWAP_STRATEGY_TRIGGER_CONFIG.value: None,
        enums.TradingSignalOrdersAttrs.TRAILING_PROFILE_TYPE.value: None,
        enums.TradingSignalOrdersAttrs.TRAILING_PROFILE.value: None,
        enums.TradingSignalOrdersAttrs.IS_ACTIVE.value: False,
        enums.TradingSignalOrdersAttrs.ACTIVE_TRIGGER_PRICE.value: 1.14,
        enums.TradingSignalOrdersAttrs.ACTIVE_TRIGGER_ABOVE.value: True,
        enums.TradingSignalOrdersAttrs.TAG.value: "hello",
        enums.TradingSignalOrdersAttrs.ORDER_ID.value: buy_limit_order.order_id,
        enums.TradingSignalOrdersAttrs.BUNDLED_WITH.value: sell_limit_order.order_id,
        enums.TradingSignalOrdersAttrs.CHAINED_TO.value: sell_limit_order.order_id,
        enums.TradingSignalOrdersAttrs.ADDITIONAL_ORDERS.value: [],
        enums.TradingSignalOrdersAttrs.ASSOCIATED_ORDER_IDS.value: ["1"],
        enums.TradingSignalOrdersAttrs.UPDATE_WITH_TRIGGERING_ORDER_FEES.value: True,
        enums.TradingSignalOrdersAttrs.CANCEL_POLICY_TYPE.value: personal_data.ChainedOrderFillingPriceOrderCancelPolicy.__name__,
        enums.TradingSignalOrdersAttrs.CANCEL_POLICY_KWARGS.value: {},
    }

    order_group = personal_data.OneCancelsTheOtherOrderGroup(
        "group_name",
        buy_limit_order.exchange_manager.exchange_personal_data.orders_manager
    )
    buy_limit_order.add_to_order_group(order_group)
    buy_limit_order.associate_to_entry("2")
    buy_limit_order.associate_to_entry("3")
    buy_limit_order.update_with_triggering_order_fees = False
    buy_limit_order.trigger_above = True
    buy_limit_order.trailing_profile = personal_data.FilledTakeProfitTrailingProfile([
        personal_data.TrailingPriceStep(price, price, True)
        for price in (10000, 12000, 13000)
    ])
    final_order_desc = {
        enums.TradingSignalCommonsAttrs.ACTION.value: enums.TradingSignalOrdersActions.CREATE.value,
        enums.TradingSignalOrdersAttrs.SIDE.value: enums.TradeOrderSide.BUY.value,
        enums.TradingSignalOrdersAttrs.STRATEGY.value: "strat",
        enums.TradingSignalOrdersAttrs.SYMBOL.value: "BTC/ETH",
        enums.TradingSignalOrdersAttrs.EXCHANGE.value: "binanceus",
        enums.TradingSignalOrdersAttrs.EXCHANGE_TYPE.value: enums.ExchangeTypes.SPOT.value,
        enums.TradingSignalOrdersAttrs.TYPE.value: enums.TraderOrderType.BUY_LIMIT.value,
        enums.TradingSignalOrdersAttrs.QUANTITY.value: float(buy_limit_order.origin_quantity),
        enums.TradingSignalOrdersAttrs.TARGET_AMOUNT.value: "10",
        enums.TradingSignalOrdersAttrs.TARGET_POSITION.value: "8%",
        enums.TradingSignalOrdersAttrs.TRIGGER_ABOVE.value: True,
        enums.TradingSignalOrdersAttrs.UPDATED_TARGET_AMOUNT.value: "10",
        enums.TradingSignalOrdersAttrs.UPDATED_TARGET_POSITION.value: "8%",
        enums.TradingSignalOrdersAttrs.LIMIT_PRICE.value: 111.545445445,
        enums.TradingSignalOrdersAttrs.UPDATED_LIMIT_PRICE.value: 111.545445445,
        enums.TradingSignalOrdersAttrs.STOP_PRICE.value: 111.1,
        enums.TradingSignalOrdersAttrs.UPDATED_STOP_PRICE.value: 111.1,
        enums.TradingSignalOrdersAttrs.CURRENT_PRICE.value: 111.0,
        enums.TradingSignalOrdersAttrs.UPDATED_CURRENT_PRICE.value: 111.0,
        enums.TradingSignalOrdersAttrs.REDUCE_ONLY.value: True,
        enums.TradingSignalOrdersAttrs.POST_ONLY.value: False,
        enums.TradingSignalOrdersAttrs.GROUP_ID.value: order_group.name,
        enums.TradingSignalOrdersAttrs.GROUP_TYPE.value: personal_data.OneCancelsTheOtherOrderGroup.__name__,
        enums.TradingSignalOrdersAttrs.ACTIVE_SWAP_STRATEGY_TYPE.value: personal_data.StopFirstActiveOrderSwapStrategy.__name__,
        enums.TradingSignalOrdersAttrs.ACTIVE_SWAP_STRATEGY_TIMEOUT.value: constants.ACTIVE_ORDER_STRATEGY_SWAP_TIMEOUT,
        enums.TradingSignalOrdersAttrs.ACTIVE_SWAP_STRATEGY_TRIGGER_CONFIG.value: enums.ActiveOrderSwapTriggerPriceConfiguration.FILLING_PRICE.value,
        enums.TradingSignalOrdersAttrs.TRAILING_PROFILE_TYPE.value:
            personal_data.FilledTakeProfitTrailingProfile.get_type().value,
        enums.TradingSignalOrdersAttrs.TRAILING_PROFILE.value: buy_limit_order.trailing_profile.to_dict(),
        enums.TradingSignalOrdersAttrs.IS_ACTIVE.value: False,
        enums.TradingSignalOrdersAttrs.ACTIVE_TRIGGER_PRICE.value: 1.14,
        enums.TradingSignalOrdersAttrs.ACTIVE_TRIGGER_ABOVE.value: True,
        enums.TradingSignalOrdersAttrs.TAG.value: "hello",
        enums.TradingSignalOrdersAttrs.ORDER_ID.value: buy_limit_order.order_id,
        enums.TradingSignalOrdersAttrs.BUNDLED_WITH.value: sell_limit_order.order_id,
        enums.TradingSignalOrdersAttrs.CHAINED_TO.value: sell_limit_order.order_id,
        enums.TradingSignalOrdersAttrs.ADDITIONAL_ORDERS.value: [],
        enums.TradingSignalOrdersAttrs.ASSOCIATED_ORDER_IDS.value: ["1", "2", "3"],
        enums.TradingSignalOrdersAttrs.UPDATE_WITH_TRIGGERING_ORDER_FEES.value: False,
        enums.TradingSignalOrdersAttrs.CANCEL_POLICY_TYPE.value: personal_data.ChainedOrderFillingPriceOrderCancelPolicy.__name__,
        enums.TradingSignalOrdersAttrs.CANCEL_POLICY_KWARGS.value: {},
    }
    assert signals.create_order_signal_content(
        buy_limit_order,
        enums.TradingSignalOrdersActions.CREATE,
        "strat",
        exchange_manager,
        target_amount="1%",
        target_position="2",
        updated_target_amount="10",
        updated_target_position="8%",
        updated_limit_price=decimal.Decimal("111.545445445"),
        updated_stop_price=decimal.Decimal("111.1"),
        updated_current_price=decimal.Decimal("111"),
    ) == final_order_desc

    # with cleared order, gives the same signal
    buy_limit_order.clear()
    assert signals.create_order_signal_content(
        buy_limit_order,
        enums.TradingSignalOrdersActions.CREATE,
        "strat",
        exchange_manager,
        target_amount="1%",
        target_position="2",
        updated_target_amount="10",
        updated_target_position="8%",
        updated_limit_price=decimal.Decimal("111.545445445"),
        updated_stop_price=decimal.Decimal("111.1"),
        updated_current_price=decimal.Decimal("111"),
    ) == final_order_desc


def test_get_order_dependency(buy_limit_order):
    assert signals.get_order_dependency(buy_limit_order) == commons_signals.SignalDependencies([
        {enums.TradingSignalDependencies.ORDER_ID.value: buy_limit_order.order_id}
    ])


def get_orders_dependencies(buy_limit_order, sell_limit_order):
    assert signals.get_orders_dependencies([buy_limit_order, sell_limit_order]) == commons_signals.SignalDependencies([
        {enums.TradingSignalDependencies.ORDER_ID.value: buy_limit_order.order_id},
        {enums.TradingSignalDependencies.ORDER_ID.value: sell_limit_order.order_id}
    ])


def test_get_position_dependency():
    position = mock.Mock(symbol="BTC/USDT:USDT")
    assert signals.get_position_dependency(position) == commons_signals.SignalDependencies([
        {enums.TradingSignalDependencies.POSITION_SYMBOL.value: "BTC/USDT:USDT"}
    ])


def test_are_dependencies_filled():
    no_dep_signal = commons_signals.Signal(
        topic=enums.TradingSignalTopics.ORDERS.value,
        content={}
    )
    assert signals_util.are_dependencies_filled(no_dep_signal, []) is True
    single_dep_signal = commons_signals.Signal(
        topic=enums.TradingSignalTopics.ORDERS.value,
        content={},
        dependencies=signals.get_order_dependency(mock.Mock(order_id="123"))
    )
    assert signals_util.are_dependencies_filled(single_dep_signal, []) is False
    dual_dep_signal = commons_signals.Signal(
        topic=enums.TradingSignalTopics.ORDERS.value,
        content={},
        dependencies=signals.get_orders_dependencies([mock.Mock(order_id="123"), mock.Mock(order_id="456")])
    )
    assert signals_util.are_dependencies_filled(dual_dep_signal, []) is False
    succeeded_signals = [
        commons_signals.Signal(
            topic=enums.TradingSignalTopics.ORDERS.value,
            content={
                enums.TradingSignalOrdersAttrs.ORDER_ID.value: "123"
            }
        )
    ]
    assert signals_util.are_dependencies_filled(no_dep_signal, succeeded_signals) is True
    # dep is filled
    assert signals_util.are_dependencies_filled(single_dep_signal, succeeded_signals) is True
    succeeded_signals.append(
        commons_signals.Signal(
            topic=enums.TradingSignalTopics.ORDERS.value,
            content={
                enums.TradingSignalOrdersAttrs.ORDER_ID.value: "456-wrong"  # not the expected id
            }
        )
    )
    assert signals_util.are_dependencies_filled(no_dep_signal, succeeded_signals) is True
    assert signals_util.are_dependencies_filled(single_dep_signal, succeeded_signals) is True
    assert signals_util.are_dependencies_filled(dual_dep_signal, succeeded_signals) is False
    succeeded_signals.append(
        commons_signals.Signal(
            topic=enums.TradingSignalTopics.ORDERS.value,
            content={
                enums.TradingSignalOrdersAttrs.ORDER_ID.value: "456"  # the last expected id
            }
        )
    )
    assert signals_util.are_dependencies_filled(no_dep_signal, succeeded_signals) is True
    assert signals_util.are_dependencies_filled(single_dep_signal, succeeded_signals) is True
    assert signals_util.are_dependencies_filled(dual_dep_signal, succeeded_signals) is True


def test_get_signals_filled_dependencies():
    assert signals_util._get_signals_filled_dependencies([]) == commons_signals.SignalDependencies()
    signal = commons_signals.Signal(
        topic=enums.TradingSignalTopics.ORDERS.value,
        content={}
    )
    assert signals_util._get_signals_filled_dependencies([signal]) == commons_signals.SignalDependencies()
    signal = commons_signals.Signal(
        topic=enums.TradingSignalTopics.ORDERS.value,
        content={
            enums.TradingSignalOrdersAttrs.ORDER_ID.value: "123"
        }
    )
    assert signals_util._get_signals_filled_dependencies([signal]) == commons_signals.SignalDependencies([
        {enums.TradingSignalDependencies.ORDER_ID.value: "123"}
    ])
    signal_2 = commons_signals.Signal(
        topic=enums.TradingSignalTopics.ORDERS.value,
        content={
            enums.TradingSignalOrdersAttrs.ORDER_ID.value: "456"
        }
    )
    signal_3 = commons_signals.Signal(
        topic=enums.TradingSignalTopics.POSITIONS.value,
        content={
            enums.TradingSignalPositionsAttrs.SYMBOL.value: "BTC/USDT:USDT"
        }
    )
    assert signals_util._get_signals_filled_dependencies([signal, signal_2, signal_3]) == commons_signals.SignalDependencies([
        {enums.TradingSignalDependencies.ORDER_ID.value: "123"},
        {enums.TradingSignalDependencies.ORDER_ID.value: "456"},
        {enums.TradingSignalDependencies.POSITION_SYMBOL.value: "BTC/USDT:USDT"}
    ])

def test_get_signal_filled_dependency():
    signal = commons_signals.Signal(
        topic=enums.TradingSignalTopics.ORDERS.value,
        content={}
    )
    assert signals_util._get_signal_filled_dependency(signal) == None
    signal = commons_signals.Signal(
        topic=enums.TradingSignalTopics.ORDERS.value,
        content={
            enums.TradingSignalOrdersAttrs.ORDER_ID.value: "123"
        }
    )
    assert signals_util._get_signal_filled_dependency(signal) == {
        enums.TradingSignalDependencies.ORDER_ID.value: "123"
    }
    
    signal = commons_signals.Signal(
        topic=enums.TradingSignalTopics.POSITIONS.value,
        content={
            enums.TradingSignalPositionsAttrs.SYMBOL.value: "BTC/USDT:USDT"
        }
    )
    assert signals_util._get_signal_filled_dependency(signal) == {
        enums.TradingSignalDependencies.POSITION_SYMBOL.value: "BTC/USDT:USDT"
    }

    with pytest.raises(NotImplementedError):
        signals_util._get_signal_filled_dependency(commons_signals.Signal(
            topic="unknown_topic",
            content={}
        ))



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.enums as enums
import octobot_trading.constants as constants

import octobot_trading.exchange_data.contracts.future_contract as future_contract


class OptionContract(future_contract.FutureContract):
    def __init__(self, pair, margin_type, 
                 contract_type: enums.OptionContractType,
                 contract_size=constants.ONE,
                 maximum_leverage=constants.ONE,
                 current_leverage=constants.ONE,
                 position_mode=enums.PositionMode.ONE_WAY,
                 maintenance_margin_rate=constants.DEFAULT_SYMBOL_MAINTENANCE_MARGIN_RATE,
                 minimum_tick_size=0.5,
                 take_profit_stop_loss_mode=None):
        super().__init__(pair, margin_type=margin_type, contract_type=contract_type, contract_size=contract_size, maximum_leverage=maximum_leverage, current_leverage=current_leverage, position_mode=position_mode, maintenance_margin_rate=maintenance_margin_rate, minimum_tick_size=minimum_tick_size, take_profit_stop_loss_mode=take_profit_stop_loss_mode)

    def is_inverse_contract(self):
        """
        Inverse Contract is a contract using the coin itself as collateral
        if not inverted the contract uses fiat as the collateral
        :return: True if the contract is an inverse contract
        """
        return self.contract_type in [enums.OptionContractType.INVERSE_EXPIRABLE]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import pytest

import octobot_trading.errors as errors
import octobot_trading.constants as constants
import octobot_trading.personal_data.portfolios.assets.spot_asset as spot_asset

ASSET_CURRENCY_NAME = "BTC"


def test___eq__():
    asset = spot_asset.SpotAsset(ASSET_CURRENCY_NAME, constants.ZERO, constants.ZERO)
    assert asset == spot_asset.SpotAsset(ASSET_CURRENCY_NAME, constants.ZERO, constants.ZERO)
    assert not asset == spot_asset.SpotAsset(ASSET_CURRENCY_NAME, constants.ONE_HUNDRED, constants.ZERO)
    assert not asset == spot_asset.SpotAsset(ASSET_CURRENCY_NAME, constants.ZERO, constants.ONE_HUNDRED)


def test_update():
    asset = spot_asset.SpotAsset(ASSET_CURRENCY_NAME, constants.ZERO, constants.ZERO)
    assert not asset.update(available=constants.ZERO, total=constants.ZERO)
    assert asset.update(available=constants.ZERO, total=decimal.Decimal(5))
    assert asset.total == decimal.Decimal(5)


def test_set():
    asset = spot_asset.SpotAsset(ASSET_CURRENCY_NAME, constants.ZERO, constants.ZERO)
    assert not asset.set(available=constants.ZERO, total=constants.ZERO)
    assert asset.set(available=decimal.Decimal(5), total=decimal.Decimal(5))
    assert not asset.set(available=decimal.Decimal(5), total=decimal.Decimal(5))
    assert asset.available == decimal.Decimal(5)
    assert asset.total == decimal.Decimal(5)


def test_restore_available():
    asset = spot_asset.SpotAsset(ASSET_CURRENCY_NAME, available=constants.ONE, total=constants.ONE_HUNDRED)
    assert not asset.available == asset.total
    asset.restore_available()
    assert asset.available == asset.total == constants.ONE_HUNDRED


def test_reset():
    asset = spot_asset.SpotAsset(ASSET_CURRENCY_NAME, available=constants.ONE_HUNDRED, total=constants.ONE_HUNDRED)
    assert asset.available == asset.total == constants.ONE_HUNDRED
    asset.reset()
    assert asset.available == asset.total == constants.ZERO


def test_restore():
    asset = spot_asset.SpotAsset(ASSET_CURRENCY_NAME, total=constants.ONE_HUNDRED, available=constants.ONE_HUNDRED)

    asset.update(total=-decimal.Decimal(50))
    assert asset.total == decimal.Decimal(50)
    with pytest.raises(errors.PortfolioNegativeValueError):
        asset.update(total=-decimal.Decimal(70))
    assert asset.total == decimal.Decimal(50)

    asset.update(available=-decimal.Decimal(2.5))
    assert asset.available == decimal.Decimal(97.5)
    asset.update(available=-decimal.Decimal(2.5))
    assert asset.available == decimal.Decimal(95)
    with pytest.raises(errors.PortfolioNegativeValueError):
        asset.update(available=-decimal.Decimal(96))
    assert asset.available == decimal.Decimal(95)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import dataclasses
import octobot_trading.personal_data.orders.order_util as order_util
import octobot_trading.personal_data.orders.cancel_policies.order_cancel_policy as order_cancel_policy_import


@dataclasses.dataclass
class ChainedOrderFillingPriceOrderCancelPolicy(order_cancel_policy_import.OrderCancelPolicy):
    """
    Will cancel the order if at least one of the chained orders filling price is reached.
    """

    def should_cancel(self, order) -> bool:
        if order.is_cleared():
            self.get_logger().error(
                f"Ignored cancel policy: order {str(order)} has been cleared"
            )
            return False
        if not order.chained_orders:
            self.get_logger().error(
                f"Ignored cancel policy: order {str(order)} has no chained orders"
            )
            return False
        current_price, up_to_date = order_util.get_potentially_outdated_price(
            order.trader.exchange_manager, order.symbol
        )
        if not up_to_date:
            self.get_logger().error(
                f"Ignored cancel policy: order {str(order)} mark price: {current_price} is outdated"
            )
            return False
        for chained_order in order.chained_orders:
            if chained_order.trigger_above and current_price > chained_order.get_filling_price():
                # this chained order price is reached
                return True
            if not chained_order.trigger_above and current_price < chained_order.get_filling_price():
                # this chained order price is reached
                return True
        # no chained order price is reached
        return False



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import typing

import octobot_commons.logging as logging
import octobot_commons.signals as signals

import octobot_trading.personal_data as personal_data
import octobot_trading.enums
import octobot_trading.constants as constants
import octobot_trading.storage as storage

LOGGER = logging.get_logger(constants.API_LOGGER_TAG)


async def create_order(exchange_manager,
                       order_type: octobot_trading.enums.TraderOrderType,
                       symbol: str,
                       current_price: float,
                       quantity: float,
                       price: float,
                       wait_for_creation=True,
                       creation_timeout=constants.INDIVIDUAL_ORDER_SYNC_TIMEOUT,
                       force_if_disabled=True
    ) -> personal_data.Order:
    return await exchange_manager.trader.create_order(
        exchange_manager.trader.create_order_instance(order_type=order_type,
                                                      symbol=symbol,
                                                      current_price=current_price,
                                                      quantity=quantity,
                                                      price=price),
        wait_for_creation=wait_for_creation,
        creation_timeout=creation_timeout,
        force_if_disabled=force_if_disabled
    )


def get_open_orders(exchange_manager, symbol=None, active=None) -> list:
    return exchange_manager.exchange_personal_data.orders_manager.get_open_orders(symbol=symbol, active=active)


def get_all_orders(exchange_manager, symbol=None, active=None) -> list:
    return exchange_manager.exchange_personal_data.orders_manager.get_all_orders(symbol=symbol, active=active)


def get_pending_creation_orders(exchange_manager) -> list:
    return exchange_manager.exchange_personal_data.orders_manager.pending_creation_orders


async def cancel_all_open_orders(
    exchange_manager, emit_trading_signals=True, force_if_disabled=True
) -> tuple[bool, list]:
    return await exchange_manager.trader.cancel_all_open_orders(
        emit_trading_signals=emit_trading_signals, force_if_disabled=force_if_disabled
    )


async def cancel_all_open_orders_with_currency(
    exchange_manager, currency, emit_trading_signals=True, dependencies=None, force_if_disabled=True
) -> tuple[bool, signals.SignalDependencies]:
    return await exchange_manager.trader.cancel_all_open_orders_with_currency(
        currency, emit_trading_signals=emit_trading_signals, dependencies=dependencies, force_if_disabled=force_if_disabled
    )


async def cancel_order_with_id(
    exchange_manager, order_id, emit_trading_signals=True, wait_for_cancelling=True, dependencies=None, force_if_disabled=True
) -> tuple[bool, typing.Optional[signals.SignalDependencies]]:
    return await exchange_manager.trader.cancel_order_with_id(
        order_id, emit_trading_signals=emit_trading_signals, wait_for_cancelling=wait_for_cancelling, dependencies=dependencies, force_if_disabled=force_if_disabled
    )


def get_order_exchange_name(order) -> str:
    return order.exchange_manager.get_exchange_name()


def order_to_dict(order) -> dict:
    return order.to_dict()


def parse_order_type(dict_order) -> octobot_trading.enums.TraderOrderType:
    return personal_data.parse_order_type(dict_order)[1]


def parse_order_status(dict_order) -> octobot_trading.enums.OrderStatus:
    return personal_data.parse_order_status(dict_order)


def is_order_pending(order: personal_data.Order) -> bool:
    return order.status == octobot_trading.enums.OrderStatus.PENDING_CREATION


def get_order_profitability(exchange_manager, order_id) -> float:
    try:
        return exchange_manager.exchange_personal_data.orders_manager.get_order(order_id).get_profitability()
    except KeyError:
        # try in trades (order might be filled and stored in trades)
        return exchange_manager.exchange_personal_data.trades_manager.get_trade(order_id).trade_profitability


def get_minimal_order_cost(exchange_manager, symbol, default_price=None) -> float:
    return personal_data.get_minimal_order_cost(
        exchange_manager.exchange.get_market_status(symbol, with_fixer=False),
        default_price=default_price
    )


def get_order_trailing_profile_dict(order: personal_data.Order) -> dict:
    return storage.get_order_trailing_profile_dict(order)


def get_orders_locked_amounts_by_asset(open_orders: list[personal_data.Order]) -> dict[str, decimal.Decimal]:
    return personal_data.get_orders_locked_amounts_by_asset(open_orders)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import mock
import pytest

import octobot_trading.personal_data.orders.cancel_policies.expiration_time_order_cancel_policy as expiration_time_order_cancel_policy_import

from tests import event_loop
from tests.exchanges import simulated_exchange_manager
from tests.exchanges.traders import trader_simulator

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


class TestExpirationTimeOrderCancelPolicy:
    """Test the ExpirationTimeOrderCancelPolicy class"""

    async def test_should_cancel_when_expiration_time_reached(self, trader_simulator):
        """Test that should_cancel returns True when expiration time is reached"""
        config, exchange_manager_inst, trader_inst = trader_simulator
        
        # Set current time to 1000
        current_time = 1000.0
        expiration_time = 1000.0  # Same as current time, should cancel
        
        exchange_mock = mock.Mock()
        exchange_mock.get_exchange_current_time.return_value = current_time
        exchange_manager_inst.exchange = exchange_mock
        
        order = mock.Mock()
        order.is_cleared.return_value = False
        order.trader = trader_inst
        order.trader.exchange_manager = exchange_manager_inst
        
        policy = expiration_time_order_cancel_policy_import.ExpirationTimeOrderCancelPolicy(
            expiration_time=expiration_time
        )
        
        assert policy.should_cancel(order) is True

    async def test_should_cancel_when_expiration_time_passed(self, trader_simulator):
        """Test that should_cancel returns True when expiration time has passed"""
        config, exchange_manager_inst, trader_inst = trader_simulator
        
        # Set current time to 2000, expiration was at 1000
        # expiration_time >= current_time: 1000 >= 2000 = False
        current_time = 2000.0
        expiration_time = 1000.0
        
        exchange_mock = mock.Mock()
        exchange_mock.get_exchange_current_time.return_value = current_time
        exchange_manager_inst.exchange = exchange_mock
        
        order = mock.Mock()
        order.is_cleared.return_value = False
        order.trader = trader_inst
        order.trader.exchange_manager = exchange_manager_inst
        
        policy = expiration_time_order_cancel_policy_import.ExpirationTimeOrderCancelPolicy(
            expiration_time=expiration_time
        )
        
        assert policy.should_cancel(order) is True

    async def test_should_not_cancel_when_expiration_time_not_reached(self, trader_simulator):
        """Test that should_cancel returns False when expiration time has not been reached (expiration in future)"""
        config, exchange_manager_inst, trader_inst = trader_simulator
        
        # Set current time to 500, expiration is at 1000
        # expiration_time >= current_time: 1000 >= 500 = True
        current_time = 500.0
        expiration_time = 1000.0
        
        exchange_mock = mock.Mock()
        exchange_mock.get_exchange_current_time.return_value = current_time
        exchange_manager_inst.exchange = exchange_mock
        
        order = mock.Mock()
        order.is_cleared.return_value = False
        order.trader = trader_inst
        order.trader.exchange_manager = exchange_manager_inst
        
        policy = expiration_time_order_cancel_policy_import.ExpirationTimeOrderCancelPolicy(
            expiration_time=expiration_time
        )
        
        assert policy.should_cancel(order) is False

    async def test_should_not_cancel_when_order_cleared(self, trader_simulator):
        """Test that should_cancel returns False when order is cleared"""
        config, exchange_manager_inst, trader_inst = trader_simulator
        
        current_time = 1000.0
        expiration_time = 1000.0
        
        exchange_mock = mock.Mock()
        exchange_mock.get_exchange_current_time.return_value = current_time
        exchange_manager_inst.exchange = exchange_mock
        
        order = mock.Mock()
        order.is_cleared.return_value = True  # Order is cleared
        order.trader = trader_inst
        order.trader.exchange_manager = exchange_manager_inst
        
        policy = expiration_time_order_cancel_policy_import.ExpirationTimeOrderCancelPolicy(
            expiration_time=expiration_time
        )
        logger_mock = mock.Mock(error=mock.Mock())
        with mock.patch.object(policy, "get_logger", mock.Mock(return_value=logger_mock)):
            assert policy.should_cancel(order) is False
            # Verify error was logged
            logger_mock.error.assert_called_once()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import mock
import pytest

import octobot_trading.personal_data as personal_data
import octobot_trading.personal_data.orders.groups.balanced_take_profit_and_stop_order_group as \
    balanced_take_profit_and_stop_order_group
import octobot_trading.constants as constants
import octobot_trading.enums as enums
import octobot_trading.personal_data.orders.order_util as order_util
from tests.exchanges import backtesting_config, backtesting_exchange_manager, fake_backtesting
from tests.personal_data.orders.groups import order_mock
from tests.personal_data import DEFAULT_ORDER_SYMBOL
from tests.personal_data.orders import created_order
from tests.exchanges import simulated_trader, simulated_exchange_manager


# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


@pytest.fixture
def btps_group(backtesting_exchange_manager):
    orders_manager = mock.Mock()
    orders_manager.get_order_from_group = mock.Mock()
    return personal_data.BalancedTakeProfitAndStopOrderGroup("name",  orders_manager)


@pytest.fixture
def side_balance():
    return balanced_take_profit_and_stop_order_group.SideBalance(None, False)


async def test_on_fill(btps_group):
    order = order_mock()
    # mock btps_group.orders_manager instead of btps_group._balance_orders to allow cythonized class tests
    with mock.patch.object(btps_group.orders_manager, "get_order_from_group", mock.Mock(return_value=[])) \
            as get_order_from_group_mock:
        await btps_group.on_fill(order)
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()

        await btps_group.on_fill(order, ["hi", "ho"])
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()


async def test_on_cancel(btps_group):
    order = order_mock()
    with mock.patch.object(btps_group.orders_manager, "get_order_from_group", mock.Mock(return_value=[])) \
            as get_order_from_group_mock:
        await btps_group.on_cancel(order)
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()

        await btps_group.on_cancel(order, ["hi", "ho"])
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()


async def test_enable(btps_group):
    with mock.patch.object(btps_group.orders_manager, "get_order_from_group", mock.Mock(return_value=[])) \
            as get_order_from_group_mock:
        await btps_group.enable(False)
        get_order_from_group_mock.assert_not_called()
        await btps_group.enable(True)
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()
        await btps_group.enable(True)
        get_order_from_group_mock.assert_called_once()


async def test_can_create_order(btps_group):
    order_1 = order_mock(origin_quantity=decimal.Decimal(1), order_type=enums.TraderOrderType.SELL_LIMIT,
                         origin_price=decimal.Decimal(1), created_last_price=decimal.Decimal(2))
    order_2 = order_mock(origin_quantity=decimal.Decimal(5), order_type=enums.TraderOrderType.BUY_MARKET,
                         origin_price=decimal.Decimal(1), created_last_price=decimal.Decimal(2))
    order_3 = order_mock(origin_quantity=decimal.Decimal(10), order_type=enums.TraderOrderType.TRAILING_STOP,
                         origin_price=decimal.Decimal(1), created_last_price=decimal.Decimal(2))
    with mock.patch.object(btps_group.orders_manager, "get_order_from_group",
                           mock.Mock(return_value=[])) as get_order_from_group_mock:
        # 0 size order
        assert btps_group.can_create_order(enums.TraderOrderType.STOP_LOSS, decimal.Decimal(0)) is True
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()
        # no order, no imbalance
        assert btps_group.can_create_order(enums.TraderOrderType.STOP_LOSS, decimal.Decimal(1)) is False
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()
    with mock.patch.object(btps_group.orders_manager, "get_order_from_group",
                           mock.Mock(return_value=[order_1, order_2, order_3])) as get_order_from_group_mock:
        # enough imbalance
        assert btps_group.can_create_order(enums.TraderOrderType.SELL_LIMIT, decimal.Decimal(1)) is True
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()
        assert btps_group.can_create_order(enums.TraderOrderType.SELL_LIMIT, decimal.Decimal(4)) is True
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()

        # not enough
        assert btps_group.can_create_order(enums.TraderOrderType.SELL_LIMIT, decimal.Decimal("4.000001")) is False
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()
        assert btps_group.can_create_order(enums.TraderOrderType.STOP_LOSS, decimal.Decimal("0.00001")) is False
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()


async def test_get_max_order_quantity(btps_group):
    order_1 = order_mock(origin_quantity=decimal.Decimal(1), order_type=enums.TraderOrderType.SELL_LIMIT,
                         origin_price=decimal.Decimal(1), created_last_price=decimal.Decimal(2))
    order_2 = order_mock(origin_quantity=decimal.Decimal(5), order_type=enums.TraderOrderType.BUY_MARKET,
                         origin_price=decimal.Decimal(1), created_last_price=decimal.Decimal(2))
    order_3 = order_mock(origin_quantity=decimal.Decimal(10), order_type=enums.TraderOrderType.TRAILING_STOP,
                         origin_price=decimal.Decimal(1), created_last_price=decimal.Decimal(2))
    with mock.patch.object(btps_group.orders_manager, "get_order_from_group",
                           mock.Mock(return_value=[])) as get_order_from_group_mock:
        # no order, no imbalance
        assert btps_group.get_max_order_quantity(enums.TraderOrderType.STOP_LOSS) == constants.ZERO
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()
    with mock.patch.object(btps_group.orders_manager, "get_order_from_group",
                           mock.Mock(return_value=[order_1, order_2, order_3])) as get_order_from_group_mock:
        # imbalance
        assert btps_group.get_max_order_quantity(enums.TraderOrderType.SELL_LIMIT) == decimal.Decimal(4)
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()
        assert btps_group.get_max_order_quantity(enums.TraderOrderType.STOP_LOSS) == constants.ZERO
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()


async def test_balance_orders(btps_group):
    order_1 = order_mock(origin_price=decimal.Decimal(1), created_last_price=decimal.Decimal(10),
                         order_type=enums.TraderOrderType.SELL_LIMIT, origin_quantity=decimal.Decimal("1.5"))
    order_2 = order_mock(origin_price=decimal.Decimal(5), created_last_price=decimal.Decimal(3),
                         order_type=enums.TraderOrderType.SELL_LIMIT, origin_quantity=decimal.Decimal("11"))
    order_3 = order_mock(origin_price=decimal.Decimal(10), created_last_price=decimal.Decimal(11),
                         order_type=enums.TraderOrderType.STOP_LOSS, origin_quantity=decimal.Decimal("10"))
    order_4 = order_mock(origin_price=decimal.Decimal(42), created_last_price=decimal.Decimal(2),
                         order_type=enums.TraderOrderType.STOP_LOSS, origin_quantity=decimal.Decimal("20"))
    with mock.patch.object(btps_group.orders_manager, "get_order_from_group",
                           mock.Mock(return_value=[order_1, order_2, order_3, order_4])) as get_order_from_group_mock:
        base_balancing_order = ["existing_order"]
        btps_group.balancing_orders = base_balancing_order
        # 1. not enabled
        btps_group.enabled = False
        await btps_group._balance_orders(order_1, ["ignored_orders"], False)
        get_order_from_group_mock.assert_not_called()
        assert btps_group.balancing_orders == ["existing_order"]
        assert btps_group.balancing_orders is base_balancing_order

        # 2. enabled
        btps_group.enabled = True
        await btps_group._balance_orders(order_4, ["ignored_orders"], True)
        get_order_from_group_mock.assert_called_once()
        get_order_from_group_mock.reset_mock()
        order_1.trader.edit_order.assert_not_called()
        order_1.trader.cancel_order.assert_called_once_with(
            order_1, ignored_order=order_4, wait_for_cancelling=True,
            cancelling_timeout=constants.INDIVIDUAL_ORDER_SYNC_TIMEOUT,
            force_if_disabled=False
        )
        order_2.trader.edit_order.assert_called_once_with(
            order_2,
            edited_quantity=decimal.Decimal(10),
            edited_price=None,
            edited_stop_price=None,
            edited_current_price=None,
            params=None
        )
        order_2.trader.cancel_order.assert_not_called()
        order_3.trader.edit_order.assert_not_called()
        order_3.trader.cancel_order.assert_not_called()
        order_4.trader.edit_order.assert_not_called()
        order_4.trader.cancel_order.assert_not_called()
        assert btps_group.balancing_orders == ["existing_order"]
        assert btps_group.balancing_orders is not base_balancing_order  # changed list during orders registration

        def _get_order_update(order, updated_quantity):
            return {
                balanced_take_profit_and_stop_order_group.BalancedTakeProfitAndStopOrderGroup.ORDER: order,
                balanced_take_profit_and_stop_order_group.BalancedTakeProfitAndStopOrderGroup.UPDATED_QUANTITY: updated_quantity,
                balanced_take_profit_and_stop_order_group.BalancedTakeProfitAndStopOrderGroup.UPDATED_PRICE: decimal.Decimal(111),
            }
        for order in (order_1, order_2, order_3, order_4):
            order.trader.edit_order.reset_mock()
            order.trader.cancel_order.reset_mock()
        # 3. with updated price
        with mock.patch.object(
            balanced_take_profit_and_stop_order_group.SideBalance, "get_order_update",
            mock.Mock(side_effect=_get_order_update)
        ) as get_order_update:
            btps_group.enabled = True
            await btps_group._balance_orders(order_4, ["ignored_orders"], True)
            get_order_from_group_mock.assert_called_once()
            get_order_from_group_mock.reset_mock()
            order_1.trader.edit_order.assert_not_called()
            order_1.trader.cancel_order.assert_called_once_with(
                order_1, ignored_order=order_4, wait_for_cancelling=True,
                cancelling_timeout=constants.INDIVIDUAL_ORDER_SYNC_TIMEOUT,
                force_if_disabled=False
            )
            order_2.trader.edit_order.assert_called_once_with(
                order_2,
                edited_quantity=decimal.Decimal(10),
                edited_price=decimal.Decimal(111),
                edited_stop_price=None,
                edited_current_price=None,
                params=None
            )
            order_2.trader.cancel_order.assert_not_called()
            order_3.trader.edit_order.assert_called_once_with(
                order_3,
                edited_quantity=None,
                edited_price=decimal.Decimal(111),
                edited_stop_price=None,
                edited_current_price=None,
                params=None
            )
            order_3.trader.cancel_order.assert_not_called()
            order_4.trader.edit_order.assert_not_called()
            order_4.trader.cancel_order.assert_not_called()
            assert btps_group.balancing_orders == ["existing_order"]
            assert btps_group.balancing_orders is not base_balancing_order  # changed list during orders registration


async def test_get_balance(btps_group):
    order_1 = order_mock(order_type="order_type", origin_price=decimal.Decimal(1),
                         created_last_price=decimal.Decimal(2))
    order_2 = order_mock(order_type="order_type", origin_price=decimal.Decimal(1),
                         created_last_price=decimal.Decimal(2))
    with mock.patch.object(order_util, "is_stop_order", mock.Mock(return_value=False)) as is_stop_order_mock:
        with mock.patch.object(btps_group.orders_manager, "get_order_from_group", mock.Mock(return_value=[])) \
             as get_order_from_group_mock:
            balance = btps_group._get_balance(order_1, None, True)
            assert balance[btps_group.TAKE_PROFIT].orders == []
            assert balance[btps_group.STOP].orders == []
            get_order_from_group_mock.assert_called_once()
            is_stop_order_mock.assert_not_called()

        with mock.patch.object(btps_group.orders_manager, "get_order_from_group", mock.Mock(return_value=[order_1, order_2])) \
             as get_order_from_group_mock:
            balance = btps_group._get_balance(order_1, None, True)
            assert balance[btps_group.TAKE_PROFIT].orders == [order_2]
            assert balance[btps_group.STOP].orders == []
            get_order_from_group_mock.assert_called_once()
            is_stop_order_mock.assert_called_once_with("order_type")
            is_stop_order_mock.reset_mock()

        with mock.patch.object(btps_group.orders_manager, "get_order_from_group",
                               mock.Mock(return_value=[order_1, order_2])) as get_order_from_group_mock:
            balance = btps_group._get_balance(order_1, [order_2], False)
            assert balance[btps_group.TAKE_PROFIT].orders == []
            assert balance[btps_group.STOP].orders == []
            get_order_from_group_mock.assert_called_once()
            is_stop_order_mock.assert_not_called()

    with mock.patch.object(order_util, "is_stop_order", mock.Mock(return_value=True)) as is_stop_order_mock:
        with mock.patch.object(btps_group.orders_manager, "get_order_from_group",
                               mock.Mock(return_value=[order_1, order_2])) as get_order_from_group_mock:
            balance = btps_group._get_balance(order_1, None, False)
            assert balance[btps_group.TAKE_PROFIT].orders == []
            assert balance[btps_group.STOP].orders == [order_2]
            get_order_from_group_mock.assert_called_once()
            is_stop_order_mock.assert_called_once_with("order_type")
            is_stop_order_mock.reset_mock()


async def test_balanced_adapt_before_order_becoming_active_and_reverse_swap(simulated_trader):
    _, exchange_manager, trader_instance = simulated_trader
    trader_instance.allow_artificial_orders = False
    trader_instance.enable_inactive_orders = True
    trader_instance.simulate = False
    exchange_manager.is_backtesting = True  # force update_order_status
    btp_group = personal_data.BalancedTakeProfitAndStopOrderGroup(
        "name", exchange_manager.exchange_personal_data.orders_manager
    )
    assert isinstance(btp_group.active_order_swap_strategy, personal_data.StopFirstActiveOrderSwapStrategy)
    stop_loss_1 = created_order(personal_data.StopLossLimitOrder, enums.TraderOrderType.STOP_LOSS,
                              trader_instance, side=enums.TradeOrderSide.SELL)
    stop_loss_1.update(
        price=decimal.Decimal(8),
        quantity=decimal.Decimal("0.8"),
        symbol=DEFAULT_ORDER_SYMBOL,
        order_type=enums.TraderOrderType.STOP_LOSS,
        group=btp_group,
        is_active=True,
        active_trigger=personal_data.create_order_price_trigger(stop_loss_1, decimal.Decimal(8), False),
    )
    stop_loss_2 = created_order(personal_data.StopLossLimitOrder, enums.TraderOrderType.STOP_LOSS,
                              trader_instance, side=enums.TradeOrderSide.SELL)
    stop_loss_2.update(
        price=decimal.Decimal(10),
        quantity=decimal.Decimal("0.2"),
        symbol=DEFAULT_ORDER_SYMBOL,
        order_type=enums.TraderOrderType.STOP_LOSS,
        group=btp_group,
        is_active=True,
        active_trigger=personal_data.create_order_price_trigger(stop_loss_2, decimal.Decimal(10), False),
    )
    sell_limit_1 = created_order(personal_data.SellLimitOrder, enums.TraderOrderType.SELL_LIMIT,
                               trader_instance, side=enums.TradeOrderSide.SELL)
    sell_limit_1.update(
        price=decimal.Decimal(20),
        quantity=decimal.Decimal("0.4"),
        symbol=DEFAULT_ORDER_SYMBOL,
        order_type=enums.TraderOrderType.SELL_LIMIT,
        group=btp_group,
        is_active=False,
        active_trigger=personal_data.create_order_price_trigger(sell_limit_1, decimal.Decimal(20), True),
    )
    sell_limit_2 = created_order(personal_data.SellLimitOrder, enums.TraderOrderType.SELL_LIMIT,
                               trader_instance, side=enums.TradeOrderSide.SELL)
    sell_limit_2.update(
        price=decimal.Decimal(30),
        quantity=decimal.Decimal("0.25"),
        symbol=DEFAULT_ORDER_SYMBOL,
        order_type=enums.TraderOrderType.SELL_LIMIT,
        group=btp_group,
        is_active=False,
        active_trigger=personal_data.create_order_price_trigger(sell_limit_2, decimal.Decimal(30), True),
    )
    sell_limit_3 = created_order(personal_data.SellLimitOrder, enums.TraderOrderType.SELL_LIMIT,
                               trader_instance, side=enums.TradeOrderSide.SELL)
    sell_limit_3.update(
        price=decimal.Decimal(40),
        quantity=decimal.Decimal("0.35"),
        symbol=DEFAULT_ORDER_SYMBOL,
        order_type=enums.TraderOrderType.SELL_LIMIT,
        group=btp_group,
        is_active=False,
        active_trigger=personal_data.create_order_price_trigger(sell_limit_3, decimal.Decimal(40), True),
    )
    portfolio = exchange_manager.exchange_personal_data.portfolio_manager.portfolio.portfolio
    assert portfolio["BTC"].available == decimal.Decimal(10)
    locked_amount = constants.ZERO
    for order in [stop_loss_1, stop_loss_2, sell_limit_1, sell_limit_2, sell_limit_3]:
        await order.initialize()
        await order.exchange_manager.exchange_personal_data.orders_manager.upsert_order_instance(order)
        if order in [stop_loss_1, stop_loss_2]:
            # stop sell limit quantity is not removed from available but stop is
            locked_amount += order.origin_quantity
        assert portfolio["BTC"].available == decimal.Decimal(10) - locked_amount
    assert stop_loss_1.is_active is True
    assert stop_loss_2.is_active is True
    assert sorted(
        order.exchange_manager.exchange_personal_data.orders_manager.get_all_orders(active=False), key=lambda x: x.origin_price
    ) ==  sorted(
        [sell_limit_1, sell_limit_2, sell_limit_3], key=lambda x: x.origin_price
    )
    async def _edit_order(*args, quantity=None, price=None, **kwargs):
        edited_stop_loss = created_order(personal_data.StopLossLimitOrder, enums.TraderOrderType.STOP_LOSS,
                                        trader_instance, side=enums.TradeOrderSide.SELL)
        edited_stop_loss.update(
            price=price or decimal.Decimal(10),
            quantity=quantity,
            symbol=DEFAULT_ORDER_SYMBOL,
            order_type=enums.TraderOrderType.STOP_LOSS,
            group=btp_group,
            is_active=True,
            active_trigger=personal_data.create_order_price_trigger(edited_stop_loss, decimal.Decimal(10), False),
        )
        order_dict = edited_stop_loss.to_dict()
        order_dict.pop(enums.ExchangeConstantsOrderColumns.ID.value)
        return order_dict


    with mock.patch.object(exchange_manager.exchange, "cancel_order", mock.AsyncMock(return_value=enums.OrderStatus.CANCELED)) as cancel_order_mock, \
            mock.patch.object(exchange_manager.exchange, "edit_order",
                              mock.AsyncMock(side_effect=_edit_order)) as edit_order_mock:
        # sell_limit_1 become active: stop_loss_2 is now inactive, stop_loss_1 is reduced and remains active
        now_maybe_partially_inactive_orders, _ = await btp_group.adapt_before_order_becoming_active(sell_limit_1)
        sell_limit_1.is_active=True    # simulate now active order
        assert now_maybe_partially_inactive_orders == [stop_loss_2, stop_loss_1]
        cancel_order_mock.assert_called_once()
        edit_order_mock.assert_called_once()
        assert stop_loss_1.is_active is True
        assert stop_loss_2.is_active is False
        for order in [stop_loss_1, stop_loss_2, sell_limit_1, sell_limit_2, sell_limit_3]:
            assert not order.is_cleared()
        # stop_loss_2 is now inactive as well, sell limit could be updated to come active (done in strategy)
        all_orders = order.exchange_manager.exchange_personal_data.orders_manager.get_all_orders()
        assert len(all_orders) == 5
        assert stop_loss_1 in all_orders and stop_loss_2 in all_orders
        assert all(order in all_orders for order in [sell_limit_1, sell_limit_2, sell_limit_3])
        inactive_orders = order.exchange_manager.exchange_personal_data.orders_manager.get_all_orders(active=False)
        assert len(inactive_orders) == 3
        assert stop_loss_2 in inactive_orders and all(order in inactive_orders for order in [sell_limit_2, sell_limit_3])
        # stop_loss_1 is reduced
        assert stop_loss_1.origin_quantity == decimal.Decimal("0.6")  # 0.8 - 0.2
        cancel_order_mock.reset_mock()
        edit_order_mock.reset_mock()

        # fill sell_limit_1: stop_loss_2 gets canceled
        await sell_limit_1.on_fill(force_fill=True)
        # stop_loss_2 is canceled
        assert sell_limit_1.is_closed()
        assert stop_loss_2.is_closed()
        assert sell_limit_1.is_cleared()
        assert stop_loss_2.is_cleared()
        all_orders = order.exchange_manager.exchange_personal_data.orders_manager.get_all_orders()
        assert len(all_orders) == 3
        assert stop_loss_1 in all_orders and stop_loss_2 not in all_orders
        assert all(order in all_orders for order in [sell_limit_2, sell_limit_3])
        inactive_orders = order.exchange_manager.exchange_personal_data.orders_manager.get_all_orders(active=False)
        assert len(inactive_orders) == 2
        assert all(order in inactive_orders for order in [sell_limit_2, sell_limit_3])
        assert stop_loss_1.is_active is True
        assert sell_limit_2.is_active is False
        assert sell_limit_3.is_active is False

        # sell_limit_2 become active: stop_loss_1 is reduced and remains active
        # stop_loss_1 is reduced again, no order is canceled
        now_maybe_partially_inactive_orders, reverse_swap = await btp_group.adapt_before_order_becoming_active(sell_limit_2)
        sell_limit_2.is_active=True    # simulate now active order
        assert now_maybe_partially_inactive_orders == [stop_loss_1]
        cancel_order_mock.assert_not_called()
        edit_order_mock.assert_called_once()
        assert stop_loss_1.is_active is True
        assert sell_limit_2.is_active is True
        assert sell_limit_3.is_active is False
        for order in [stop_loss_1, sell_limit_2, sell_limit_3]:
            assert not order.is_cleared()
        # stop_loss_2 is now inactive as well, could be to update sell limit to come active (done in strategy)
        inactive_orders = order.exchange_manager.exchange_personal_data.orders_manager.get_all_orders(active=False)
        assert inactive_orders == [sell_limit_3]
        # stop_loss_1 is reduced
        assert stop_loss_1.origin_quantity == decimal.Decimal("0.35")  # 0.8 - 0.2 - 0.25
        cancel_order_mock.reset_mock()
        edit_order_mock.reset_mock()

        # now reverse swap
        with mock.patch.object(btp_group, "adapt_before_order_becoming_active", mock.AsyncMock()) as adapt_before_order_becoming_active_mock, \
            mock.patch.object(order_util, "create_as_active_order_on_exchange", mock.AsyncMock()) as create_as_active_order_on_exchange_mock, \
            mock.patch.object(btp_group, "_apply_update_order_actions", mock.AsyncMock()) as _apply_update_order_actions_mock, \
            mock.patch.object(order_util, "update_order_as_inactive_on_exchange", mock.AsyncMock()) as update_order_as_inactive_on_exchange_mock:
            await reverse_swap("order_0", ["order_1", "order_2"])
            update_order_as_inactive_on_exchange_mock.assert_called_once_with("order_0", False)
            adapt_before_order_becoming_active_mock.assert_not_called()
            create_as_active_order_on_exchange_mock.assert_not_called()   # no order to re-create (no cancelled order in previous step)
            _apply_update_order_actions_mock.assert_called_once()
            reverse_actions = _apply_update_order_actions_mock.mock_calls[0].args[0]
            assert len(reverse_actions) == 1
            assert reverse_actions[0][btp_group.ORDER] is stop_loss_1
            assert reverse_actions[0][btp_group.UPDATED_QUANTITY] == decimal.Decimal("0.60")

        # without deactivated orders
        await reverse_swap(sell_limit_2, [])
        # stop loss gets increased again and sell_limit_2 is back to inactive
        cancel_order_mock.assert_called_once()
        edit_order_mock.assert_called_once()
        assert stop_loss_1.is_active is True
        assert stop_loss_1.origin_quantity == decimal.Decimal("0.6")  # 0.8 - 0.2
        assert sell_limit_2.is_active is False


async def test_SideBalance_add_order(side_balance):
    assert side_balance.orders == []
    order_1 = order_mock(origin_price=decimal.Decimal(1), created_last_price=decimal.Decimal(2))
    order_2 = order_mock(origin_price=decimal.Decimal(15), created_last_price=decimal.Decimal(3))
    order_3 = order_mock(origin_price=decimal.Decimal(50), created_last_price=decimal.Decimal(55))
    side_balance.add_order(order_1)
    assert side_balance.orders == [order_1]
    side_balance.add_order(order_2)
    assert side_balance.orders == [order_2, order_1]
    side_balance.add_order(order_3)
    assert side_balance.orders == [order_2, order_3, order_1]


async def test_SideBalance_get_actions_to_balance_special_inputs(side_balance):
    with mock.patch.object(side_balance, "get_balance", mock.Mock(return_value=decimal.Decimal("-1"))):
        # target balance can't be negative (order quantity is always positive). value treated as abs()
        assert side_balance.get_actions_to_balance(constants.ONE) == {
            personal_data.BalancedTakeProfitAndStopOrderGroup.CANCEL: [],
            personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATE: [],
        }
    with mock.patch.object(side_balance, "get_balance", mock.Mock(return_value=constants.ONE)):
        # target balance == balance: nothing to do
        assert side_balance.get_actions_to_balance(constants.ONE) == {
            personal_data.BalancedTakeProfitAndStopOrderGroup.CANCEL: [],
            personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATE: [],
        }
    with mock.patch.object(side_balance, "get_balance", mock.Mock(return_value=constants.ONE)):
        # target balance > balance: nothing to do (can't create order to increase)
        assert side_balance.get_actions_to_balance(decimal.Decimal("0.6")) == {
            personal_data.BalancedTakeProfitAndStopOrderGroup.CANCEL: [],
            personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATE: [],
        }
    with mock.patch.object(side_balance, "get_balance", mock.Mock(return_value=constants.ONE)):
        # no order to create actions
        assert side_balance.get_actions_to_balance(decimal.Decimal("10")) == {
            personal_data.BalancedTakeProfitAndStopOrderGroup.CANCEL: [],
            personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATE: [],
        }


async def test_SideBalance_get_actions_to_balance_normal_inputs(side_balance):
    order_0_1 = order_mock(origin_quantity=decimal.Decimal("2.899999"))
    order_0_2 = order_mock(origin_quantity=decimal.Decimal("0.111111"))
    order_0_3 = order_mock(origin_quantity=decimal.Decimal("3.15545145441"))
    side_balance.orders = [order_0_1, order_0_2, order_0_3]
    assert side_balance.get_actions_to_balance(decimal.Decimal("3.15545145441")) == {
        personal_data.BalancedTakeProfitAndStopOrderGroup.CANCEL: [order_0_1, order_0_2],
        personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATE: [{
            personal_data.BalancedTakeProfitAndStopOrderGroup.ORDER: order_0_3,
            personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATED_QUANTITY: None,
            personal_data.BalancedTakeProfitAndStopOrderGroup.INITIAL_QUANTITY: decimal.Decimal('3.15545145441')
        }],
    }
    order_1 = order_mock(origin_quantity=decimal.Decimal("1"))
    order_2 = order_mock(origin_quantity=decimal.Decimal("6.42"))
    side_balance.orders = [order_2]
    assert side_balance.get_actions_to_balance(decimal.Decimal("3")) == {
        personal_data.BalancedTakeProfitAndStopOrderGroup.CANCEL: [],
        personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATE: [{
            personal_data.BalancedTakeProfitAndStopOrderGroup.ORDER: order_2,
            personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATED_QUANTITY: decimal.Decimal("3"),
            personal_data.BalancedTakeProfitAndStopOrderGroup.INITIAL_QUANTITY: decimal.Decimal("6.42")
        }],
    }
    side_balance.orders = [order_1, order_2]
    assert side_balance.get_actions_to_balance(decimal.Decimal("3")) == {
        personal_data.BalancedTakeProfitAndStopOrderGroup.CANCEL: [order_1],
        personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATE: [{
            personal_data.BalancedTakeProfitAndStopOrderGroup.ORDER: order_2,
            personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATED_QUANTITY: decimal.Decimal("3"),
            personal_data.BalancedTakeProfitAndStopOrderGroup.INITIAL_QUANTITY: decimal.Decimal("6.42")
        }],
    }
    order_2 = order_mock(origin_quantity=decimal.Decimal("6.42"))
    order_3 = order_mock(origin_quantity=decimal.Decimal("11.888998327557457"))
    side_balance.orders = [order_1, order_2, order_3]
    assert side_balance.get_actions_to_balance(decimal.Decimal("3")) == {
        personal_data.BalancedTakeProfitAndStopOrderGroup.CANCEL: [order_1, order_2],
        personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATE: [{
            personal_data.BalancedTakeProfitAndStopOrderGroup.ORDER: order_3,
            personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATED_QUANTITY: decimal.Decimal("3"),
            personal_data.BalancedTakeProfitAndStopOrderGroup.INITIAL_QUANTITY: decimal.Decimal("11.888998327557457")
        }],
    }
    order_4 = order_mock(origin_quantity=decimal.Decimal("0.1"))
    side_balance.orders = [order_1, order_2, order_3, order_4]
    # keep order_4 untouched as it is the closest from price and can be included in target balance
    assert side_balance.get_actions_to_balance(decimal.Decimal("3")) == {
        personal_data.BalancedTakeProfitAndStopOrderGroup.CANCEL: [order_1, order_2],
        personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATE: [{
            personal_data.BalancedTakeProfitAndStopOrderGroup.ORDER: order_3,
            personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATED_QUANTITY: decimal.Decimal("2.9"),
            personal_data.BalancedTakeProfitAndStopOrderGroup.INITIAL_QUANTITY: decimal.Decimal("11.888998327557457")
        },{
            personal_data.BalancedTakeProfitAndStopOrderGroup.ORDER: order_4,
            personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATED_QUANTITY: None,
            personal_data.BalancedTakeProfitAndStopOrderGroup.INITIAL_QUANTITY: decimal.Decimal("0.1")
        }],
    }
    side_balance.orders = [order_1, order_4, order_2, order_3]
    assert side_balance.get_actions_to_balance(decimal.Decimal("3")) == {
        personal_data.BalancedTakeProfitAndStopOrderGroup.CANCEL: [order_1, order_4, order_2],
        personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATE: [{
            personal_data.BalancedTakeProfitAndStopOrderGroup.ORDER: order_3,
            personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATED_QUANTITY: decimal.Decimal("3"),
            personal_data.BalancedTakeProfitAndStopOrderGroup.INITIAL_QUANTITY: decimal.Decimal("11.888998327557457")
        }],
    }

    side_balance.orders = [order_1, order_4, order_2, order_3]
    assert side_balance.get_actions_to_balance(constants.ZERO) == {
        personal_data.BalancedTakeProfitAndStopOrderGroup.CANCEL: [order_1, order_4, order_2, order_3],
        personal_data.BalancedTakeProfitAndStopOrderGroup.UPDATE: [],
    }


async def test_SideBalance_get_balance(side_balance):
    assert side_balance.get_balance() is constants.ZERO
    order_1 = order_mock(origin_quantity=decimal.Decimal(1))
    order_2 = order_mock(origin_quantity=decimal.Decimal(10))
    order_mock(origin_quantity=decimal.Decimal(50))
    side_balance.orders = [order_1, order_2]
    assert side_balance.get_balance() == decimal.Decimal(11)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.personal_data.portfolios.portfolio_value_holder as portfolio_value_holder
import octobot_trading.personal_data.portfolios.holders.futures_portfolio_value_holder as futures_portfolio_value_holder
import octobot_trading.personal_data.portfolios.holders.option_portfolio_value_holder as option_portfolio_value_holder
import octobot_trading.personal_data.portfolios.holders.margin_portfolio_value_holder as margin_portfolio_value_holder
import octobot_trading.personal_data.portfolios.holders.spot_portfolio_value_holder as spot_portfolio_value_holder


def create_portfolio_value_holder(exchange_manager, portfolio_manager) -> portfolio_value_holder.PortfolioValueHolder:
    """
    Create a portfolio value holder from an exchange manager and a portfolio manager
    :param exchange_manager: the exchange manager related to the new portfolio value holder instance
    :param portfolio_manager: the portfolio manager related to the new portfolio value holder instance
    :return: the created portfolio value holder instance
    """
    if exchange_manager.is_future:
        return futures_portfolio_value_holder.FuturesPortfolioValueHolder(portfolio_manager)
    if exchange_manager.is_option:
        return option_portfolio_value_holder.OptionPortfolioValueHolder(portfolio_manager)
    if exchange_manager.is_margin:
        return margin_portfolio_value_holder.MarginPortfolioValueHolder(portfolio_manager)
    return spot_portfolio_value_holder.SpotPortfolioValueHolder(portfolio_manager)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import octobot_commons.constants as commons_constants
import octobot_commons.enums as commons_enums
import octobot_trading.modes.script_keywords.basic_keywords.user_inputs as user_inputs
import octobot_trading.enums as enums
import octobot_trading.errors as errors
import octobot_trading.constants as constants


async def user_select_leverage(
    ctx,
    def_val=1,
    order=None,
    name=constants.CONFIG_LEVERAGE
):
    return await user_inputs.user_input(ctx, name, commons_enums.UserInputTypes.INT.value, def_val, order=order)


async def user_select_emit_trading_signals(ctx, identifier, def_val=False) -> bool:
    if is_emitting_signals := await user_inputs.user_input(ctx, commons_constants.CONFIG_EMIT_TRADING_SIGNALS,
                                                           commons_enums.UserInputTypes.BOOLEAN.value, def_val,
                                                           show_in_summary=False, show_in_optimizer=False):
        await user_inputs.user_input(ctx, commons_constants.CONFIG_TRADING_SIGNALS_STRATEGY,
                                     commons_enums.UserInputTypes.TEXT.value, identifier,
                                     show_in_summary=False, show_in_optimizer=False)
    return is_emitting_signals


async def set_leverage(ctx, leverage, side=None):
    if ctx.exchange_manager.is_future:
        try:
            await ctx.exchange_manager.trader.set_leverage(ctx.symbol, side, decimal.Decimal(str(leverage)))
        except errors.ContractExistsError as e:
            ctx.logger.debug(str(e))
        except NotImplementedError as e:
            ctx.logger.exception(e, True, str(e))
        except Exception as e:
            contract = ctx.exchange_manager.exchange.get_pair_future_contract(ctx.symbol)
            ctx.logger.exception(
                e,
                True,
                f"Impossible to set leverage to {leverage}, using current "
                f"value instead: {contract.current_leverage} ({e})"
            )


async def set_partial_take_profit_stop_loss(ctx, tp_sl_mode=enums.TakeProfitStopLossMode.PARTIAL.value):
    if ctx.exchange_manager.is_future:
        await ctx.exchange_manager.trader.set_symbol_take_profit_stop_loss_mode(
            ctx.symbol, enums.TakeProfitStopLossMode(tp_sl_mode)
        )



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.personal_data.portfolios.holders import futures_portfolio_value_holder
from octobot_trading.personal_data.portfolios.holders import option_portfolio_value_holder
from octobot_trading.personal_data.portfolios.holders import spot_portfolio_value_holder

from octobot_trading.personal_data.portfolios.holders.futures_portfolio_value_holder import (
    FuturesPortfolioValueHolder,
)
from octobot_trading.personal_data.portfolios.holders.option_portfolio_value_holder import (
    OptionPortfolioValueHolder,
)
from octobot_trading.personal_data.portfolios.holders.margin_portfolio_value_holder import (
    MarginPortfolioValueHolder,
)
from octobot_trading.personal_data.portfolios.holders.spot_portfolio_value_holder import (
    SpotPortfolioValueHolder,
)

__all__ = [
    "FuturesPortfolioValueHolder",
    "OptionPortfolioValueHolder",
    "MarginPortfolioValueHolder",
    "SpotPortfolioValueHolder",
]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import octobot_trading.constants as constants
import octobot_trading.personal_data.portfolios.assets.option_asset as option_asset

ASSET_CURRENCY_NAME = "BTC"


def test_str():
    asset = option_asset.OptionAsset(ASSET_CURRENCY_NAME, constants.ZERO, constants.ZERO)
    result = str(asset)
    assert "OptionAsset" in result
    assert ASSET_CURRENCY_NAME in result
    assert "Available: 0.0" in result
    assert "Total: 0.0" in result
    assert "Initial Margin: 0.0" in result
    assert "Wallet Balance: 0.0" in result
    assert "Unrealized PNL: 0.0" in result
    assert "Order Margin: 0.0" in result
    assert "Position Margin: 0.0" in result

    asset = option_asset.OptionAsset(ASSET_CURRENCY_NAME,
                                     available=decimal.Decimal(10),
                                     total=decimal.Decimal(100),
                                     initial_margin=decimal.Decimal(20),
                                     wallet_balance=decimal.Decimal(80),
                                     position_margin=decimal.Decimal(15),
                                     order_margin=decimal.Decimal(5),
                                     unrealized_pnl=decimal.Decimal(25))
    result = str(asset)
    assert "OptionAsset" in result
    assert ASSET_CURRENCY_NAME in result
    assert "Available: 10.0" in result
    assert "Total: 100.0" in result
    assert "Initial Margin: 20.0" in result
    assert "Wallet Balance: 80.0" in result
    assert "Unrealized PNL: 25.0" in result
    assert "Order Margin: 5.0" in result
    assert "Position Margin: 15.0" in result



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import dataclasses
import decimal
import typing

import octobot_trading.errors
import octobot_trading.personal_data.orders.trailing_profiles.trailing_profile_types as trailing_profile_types
import octobot_trading.personal_data.orders.trailing_profiles.trailing_profile as trailing_profile
import octobot_trading.personal_data.orders.trailing_profiles.trailing_price_step as trailing_price_step


@dataclasses.dataclass
class FilledTakeProfitTrailingProfile(trailing_profile.TrailingProfile):
    """
    To be used alongside TrailingOnFilledTPBalancedOrderGroup
    """
    steps: list[trailing_price_step.TrailingPriceStep]

    # pylint: disable=E1134
    def __post_init__(self):
        if self.steps and isinstance(self.steps[0], dict):
            self.steps = [
                trailing_price_step.TrailingPriceStep.from_dict(step)
                for step in self.steps
            ]

    @staticmethod
    def get_type() -> trailing_profile_types.TrailingProfileTypes:
        return trailing_profile_types.TrailingProfileTypes.FILLED_TAKE_PROFIT

    def update_and_get_trailing_price(self, updated_price: decimal.Decimal) -> typing.Optional[decimal.Decimal]:
        if updated_step := self._get_updated_step(updated_price):
            updated_step.reached = True
            return decimal.Decimal(str(updated_step.trailing_price))
        return None

    def _reach_outdated_steps_and_check_price(self, updated_price: decimal.Decimal) -> bool:
        ordered_steps = self._get_ordered_remaining_steps()
        first_newly_reached_step_index = None
        last_newly_reached_step_index = None
        float_price = float(updated_price)
        for index, step in enumerate(ordered_steps):
            if not step.reached and step.should_be_reached(float_price):
                if first_newly_reached_step_index is None:
                    first_newly_reached_step_index = index
                last_newly_reached_step_index = index
        if first_newly_reached_step_index is None:
            # price not reached
            return False
        if last_newly_reached_step_index > first_newly_reached_step_index:
            for step in ordered_steps[first_newly_reached_step_index:last_newly_reached_step_index]:
                # step has been reached already and should be skipped
                step.reached = True
        return True

    def _get_ordered_remaining_steps(self) -> list[trailing_price_step.TrailingPriceStep]:
        return [
            step
            for step in sorted(self.steps, key=lambda x: x.step_price * (1 if x.trigger_above else -1))
            if not step.reached
        ]

    def _get_updated_step(self, updated_price: decimal.Decimal) -> typing.Optional[trailing_price_step.TrailingPriceStep]:
        if self._reach_outdated_steps_and_check_price(updated_price):
            try:
                return self._get_ordered_remaining_steps()[0]
            except IndexError:
                raise octobot_trading.errors.ExhaustedTrailingProfileError(
                    f"All {len(self.steps)} trailing steps have been reached"
                )
        return None




# pylint: disable=W0237
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import typing

import async_channel.constants as channel_constants
import async_channel.enums as channel_enums
import octobot_commons.signals as commons_signals

import octobot_trading.exchange_channel as exchanges_channel
import octobot_trading.constants as constants


class ModeChannelConsumer(exchanges_channel.ExchangeChannelInternalConsumer):
    pass


class ModeChannelProducer(exchanges_channel.ExchangeChannelProducer):
    async def send(self,
                   final_note=constants.ZERO,
                   trading_mode_name=channel_constants.CHANNEL_WILDCARD,
                   state=channel_constants.CHANNEL_WILDCARD,
                   cryptocurrency=channel_constants.CHANNEL_WILDCARD,
                   symbol=channel_constants.CHANNEL_WILDCARD,
                   time_frame=None,
                   data=None,
                   dependencies: typing.Optional[commons_signals.SignalDependencies] = None):
        for consumer in self.channel.get_filtered_consumers(trading_mode_name=trading_mode_name,
                                                            state=state,
                                                            cryptocurrency=cryptocurrency,
                                                            symbol=symbol,
                                                            time_frame=time_frame):
            await consumer.queue.put({
                "final_note": final_note,
                "state": state,
                "trading_mode_name": trading_mode_name,
                "cryptocurrency": cryptocurrency,
                "symbol": symbol,
                "time_frame": time_frame,
                "data": data,
                "dependencies": dependencies
            })


class ModeChannel(exchanges_channel.ExchangeChannel):
    PRODUCER_CLASS = ModeChannelProducer
    CONSUMER_CLASS = ModeChannelConsumer
    DEFAULT_PRIORITY_LEVEL = channel_enums.ChannelConsumerPriorityLevels.MEDIUM.value

    TRADING_MODE_NAME_KEY = "trading_mode_name"
    STATE_KEY = "state"
    CRYPTOCURRENCY_KEY = "cryptocurrency"
    SYMBOL_KEY = "symbol"
    TIME_FRAME_KEY = "time_frame"

    async def new_consumer(self,
                           callback: object = None,  # shouldn't be provided here (InternalConsumer)
                           consumer_instance: ModeChannelConsumer = None,
                           trading_mode_name: str = channel_constants.CHANNEL_WILDCARD,
                           state=channel_constants.CHANNEL_WILDCARD,
                           cryptocurrency: str = channel_constants.CHANNEL_WILDCARD,
                           symbol: str = channel_constants.CHANNEL_WILDCARD,
                           time_frame=None):
        await self._add_new_consumer_and_run(consumer_instance,
                                             trading_mode_name=trading_mode_name,
                                             state=state,
                                             cryptocurrency=cryptocurrency,
                                             symbol=symbol,
                                             time_frame=time_frame)

    def get_filtered_consumers(self,
                               trading_mode_name=channel_constants.CHANNEL_WILDCARD,
                               state=channel_constants.CHANNEL_WILDCARD,
                               cryptocurrency=channel_constants.CHANNEL_WILDCARD,
                               symbol=channel_constants.CHANNEL_WILDCARD,
                               time_frame=channel_constants.CHANNEL_WILDCARD):
        return self.get_consumer_from_filters({
            self.TRADING_MODE_NAME_KEY: trading_mode_name,
            self.STATE_KEY: state,
            self.CRYPTOCURRENCY_KEY: cryptocurrency,
            self.SYMBOL_KEY: symbol,
            self.TIME_FRAME_KEY: time_frame
        })

    async def _add_new_consumer_and_run(self, consumer,
                                        trading_mode_name=channel_constants.CHANNEL_WILDCARD,
                                        state=channel_constants.CHANNEL_WILDCARD,
                                        cryptocurrency=channel_constants.CHANNEL_WILDCARD,
                                        symbol=channel_constants.CHANNEL_WILDCARD,
                                        time_frame=None):
        consumer_filters: dict = {
            self.TRADING_MODE_NAME_KEY: trading_mode_name,
            self.STATE_KEY: state,
            self.CRYPTOCURRENCY_KEY: cryptocurrency,
            self.SYMBOL_KEY: symbol
        }

        if time_frame:
            consumer_filters[self.TIME_FRAME_KEY] = time_frame

        self.add_new_consumer(consumer, consumer_filters)
        await consumer.run()
        self.logger.debug(f"Consumer started for : "
                          f"[trading_mode_name={trading_mode_name},"
                          f" state={state},"
                          f" cryptocurrency={cryptocurrency},"
                          f" symbol={symbol},"
                          f" time_frame={time_frame}]")



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import dataclasses
import typing


@dataclasses.dataclass
class ExchangeCredentialsData:
    # CEXes
    api_key: typing.Optional[str] = None
    secret: typing.Optional[str] = None
    password: typing.Optional[str] = None
    uid: typing.Optional[str] = None
    # Oauth
    auth_token: typing.Optional[str] = None
    auth_token_header_prefix: typing.Optional[str] = None
    # DEXes
    wallet_address: typing.Optional[str] = None
    private_key: typing.Optional[str] = None

    def has_credentials(self) -> bool:
        return bool(
            (self.api_key and self.secret)  # CEX
            or (self.wallet_address and self.private_key)  # DEX
        )




# pylint: disable=E0611
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import time

import octobot_commons.html_util as html_util

import octobot_trading.errors as errors
import octobot_trading.constants as constants
import octobot_trading.exchange_data.markets.channel.markets as markets_channel


class MarketsUpdater(markets_channel.MarketsProducer):
    CHANNEL_NAME = constants.MARKETS_CHANNEL
    SPOT_MARKETS_REFRESH_TIME = 86400 # 1 time per day
    FUTURES_MARKETS_REFRESH_TIME = 86400 # 1 time per day
    OPTION_MARKETS_REFRESH_TIME = 150 # every 2.5 minutes

    def __init__(self, channel):
        super().__init__(channel)
        self.refresh_time = self._get_refresh_time()
        self.tasks = []

    async def start(self):
        self.tasks = [asyncio.create_task(self.market_watcher())]

    def _get_refresh_time(self):
        if self.channel.exchange_manager.is_future:
            return self.FUTURES_MARKETS_REFRESH_TIME
        elif self.channel.exchange_manager.is_option:
            return self.OPTION_MARKETS_REFRESH_TIME
        return self.SPOT_MARKETS_REFRESH_TIME

    async def market_watcher(self):
        """
        Manage markets data refreshing for all symbols
        """
        while not self.should_stop: # As there is not markets channel consumer, we shouldn't check self.channel.is_paused for now
            try:
                started_time = time.time()
                markets = await self.channel.exchange_manager.exchange.refresh_markets()
                await self.push(markets)
                sleep_time = max(self.refresh_time - (time.time() - started_time), 0)
                await asyncio.sleep(sleep_time)
            except errors.FailedRequest as e:
                self.logger.warning(str(e))
                # avoid spamming on disconnected situation
                await asyncio.sleep(constants.FAILED_NETWORK_REQUEST_RETRY_ATTEMPTS)
            except errors.NotSupported:
                self.logger.warning(f"{self.channel.exchange_manager.exchange_name} is not supporting updates")
                await self.pause()
            except Exception as e:
                self.logger.exception(
                    e,
                    True,
                    f"Failed to update markets data : {html_util.get_html_summary_if_relevant(e)}"
                )

    async def resume(self) -> None:
        await super().resume()
        if not self.is_running:
            for task in self.tasks:
                task.cancel()
            await self.run()

    async def stop(self) -> None:
        await super().stop()
        for task in self.tasks:
            task.cancel()
        self.tasks = []



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import octobot_trading.constants as constants
import octobot_trading.personal_data.portfolios.assets.option_asset as option_asset
from octobot_trading.personal_data import OptionPortfolio


def test_create_currency_asset():
    portfolio = OptionPortfolio("test_exchange")
    
    # Test with default values
    asset = portfolio.create_currency_asset("BTC")
    assert isinstance(asset, option_asset.OptionAsset)
    assert asset.name == "BTC"
    assert asset.available == constants.ZERO
    assert asset.total == constants.ZERO
    
    # Test with custom values
    asset = portfolio.create_currency_asset("USDT", available=decimal.Decimal(100), total=decimal.Decimal(200))
    assert isinstance(asset, option_asset.OptionAsset)
    assert asset.name == "USDT"
    assert asset.available == decimal.Decimal(100)
    assert asset.total == decimal.Decimal(200)
    
    # Test with different currency
    asset = portfolio.create_currency_asset("ETH", available=decimal.Decimal(5.5), total=decimal.Decimal(10.5))
    assert isinstance(asset, option_asset.OptionAsset)
    assert asset.name == "ETH"
    assert asset.available == decimal.Decimal(5.5)
    assert asset.total == decimal.Decimal(10.5)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import re

import octobot_trading.modes.script_keywords.dsl.values as dsl_values


# Do not compile regex to use builtin regex cache
# (re always looks into cache 1st and compile regex are not cached)
QUANTITY_REGEX = r"([+-]?(?:0|[1-9]\d*)(?:\.\d*)?(?:[eE][+\-]?\d+)?)"


def parse_quantity(input_offset) -> (dsl_values.QuantityType, decimal.Decimal):
    input_offset = "" if input_offset is None else input_offset
    input_offset = str(input_offset)
    value = None
    try:
        quantity_type, value = dsl_values.QuantityType.parse(re.sub(QUANTITY_REGEX, "", input_offset))
        if input_offset == value:
            offset_value = None
        else:
            # skip fist char before replacement when 1st char is not re quantity_type value to
            # avoid replacing a part of the number ex:
            # "1e+9e" => should only replace the last "e"
            # "e1e+9" => should only replace the first "e"
            if value:
                quantity_type_pattern_index = input_offset.find(value)
                if quantity_type_pattern_index > 0 and quantity_type_pattern_index + len(value) < len(input_offset):
                    # "1e+9e" - like case
                    quantity_type_pattern_index = input_offset.rfind(value)
                if quantity_type_pattern_index >= 0:
                    input_offset = f"{input_offset[:quantity_type_pattern_index]}" \
                                   f"{input_offset[quantity_type_pattern_index + len(value):]}"
            offset_value = decimal.Decimal(input_offset)
        return quantity_type, offset_value
    except decimal.InvalidOperation as e:
        raise RuntimeError(f"Can't parse {input_offset.replace(value, '')} as decimal.Decimal ({e})") from e
    except ValueError:
        return dsl_values.QuantityType.UNKNOWN, None



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import copy
import decimal

import octobot_commons.constants as common_constants
import octobot_commons.logging as logging
import octobot_commons.asyncio_tools as asyncio_tools
import octobot_trading.constants as constants
import octobot_trading.enums as enums
import octobot_trading.personal_data as personal_data


class Portfolio:
    """
    The Portfolio class manage an exchange portfolio
    This will begin by loading current exchange portfolio (by pulling user data)
    In case of simulation this will load the CONFIG_STARTING_PORTFOLIO
    This class also manage the availability of each currency in the portfolio:
    - When an order is created it will subtract the quantity of the total
    - When an order is filled or canceled restore the availability with the real quantity
    """

    def __init__(self, exchange_name, is_simulated=False):
        super().__init__()
        self._exchange_name: str = exchange_name
        self._is_simulated: bool = is_simulated

        self.logger: logging.BotLogger = logging.get_logger(
            f"{self.__class__.__name__}{'Simulator' if is_simulated else ''}[{exchange_name}]")
        self.lock: asyncio_tools.RLock = asyncio_tools.RLock()
        self.portfolio: dict[str, personal_data.Asset] = {}
        self.reset()

    def __copy__(self):
        """
        Copy the portfolio instance
        :return: the copied portfolio object
        """
        new_portfolio = self.__class__(self._exchange_name, is_simulated=self._is_simulated)
        new_portfolio.portfolio = copy.deepcopy(self.portfolio)
        return new_portfolio

    def __eq__(self, other):
        """
        Compare two portfolio instances
        :param other: the portfolio instance to compare
        :return: True if the two portfolio instances are equals
        """
        return self.portfolio == other.portfolio

    def __str__(self):
        """
        :return: the portfolio instance str representation
        """
        return f"{personal_data.portfolio_to_float(self.portfolio)}"

    def reset(self):
        """
        Reset the portfolio dictionary
        """
        self.portfolio = {}

    def update_portfolio_from_balance(self, balance, force_replace=True):
        """
        Update portfolio from a balance dict
        :param balance: the portfolio dict
        :param force_replace: force to update portfolio. Should be False when using deltas
        :return: True if the portfolio has been updated
        """
        if force_replace:
            self.portfolio = {
                currency: self._parse_raw_currency_asset(currency=currency, raw_currency_balance=balance[currency])
                for currency in balance}
            self.logger.debug(f"Portfolio updated | {constants.CURRENT_PORTFOLIO_STRING} {self}")
            return True
        if any(
                self._update_raw_currency_asset(currency=currency, raw_currency_balance=balance[currency])
                for currency in balance
        ):
            self.logger.debug(f"Portfolio partially updated | {constants.CURRENT_PORTFOLIO_STRING} {self}")
            return True
        return False

    def get_currency_portfolio(self, currency):
        """
        Get specified currency asset from portfolio
        :param currency: the currency to get
        :return: the currency portfolio asset instance
        """
        try:
            return self.portfolio[currency]
        except KeyError:
            self.portfolio[currency] = self.create_currency_asset(currency)
            return self.portfolio[currency]

    def create_currency_asset(self, currency, available=constants.ZERO, total=constants.ZERO):
        """
        Create the currency asset instance
        :param currency: the currency name
        :param available: the available value
        :param total: the total value
        :return: the currency asset instance
        """
        raise NotImplementedError("create_currency_asset is not implemented")

    def update_portfolio_data_from_order(self, order):
        """
        Call update_portfolio_data for order currency and market
        :param order: the order that updated the portfolio
        """
        raise NotImplementedError("update_portfolio_data_from_order is not implemented")

    def update_portfolio_data_from_withdrawal(self, amount, currency):
        """
        Call update_portfolio_data for order currency and market
        :param amount: the withdrawal amount
        :param currency: the withdrawal currency
        """
        raise NotImplementedError("update_portfolio_data_from_withdrawal is not implemented")

    def update_portfolio_available_from_order(self, order, is_new_order=True):
        """
        Realize portfolio availability update
        :param order: the order that triggers the portfolio update
        :param is_new_order: True when the order is being created, False when cancelled
        """
        raise NotImplementedError("update_portfolio_available_from_order is not implemented")

    def update_portfolio_from_filled_order(self, order):
        """
        update_portfolio performs the update of the total / available quantity of a currency
        It is called only when an order is filled to update the real quantity of the currency to be set in "total" field
        :param order: the order to be taken into account
        """
        # stop losses and take profits aren't using available portfolio
        # sync available portfolio funds when order type is stop loss or take profit
        if _should_reduce_available_assets_on_fill(order):
            self.update_portfolio_available_from_order(order)

        self.update_portfolio_data_from_order(order)
        self.log_portfolio_update_from_order(order)

    def update_portfolio_from_withdrawal(self, amount, currency):
        """
        update_portfolio performs the update of the total / available quantity of a currency
        It is called when a withdrawal is requested
        :param order: the order to be taken into account
        """

        self.update_portfolio_data_from_withdrawal(amount, currency)
        self.log_portfolio_update_from_withdrawal(amount, currency)

    def update_portfolio_available(self, order, is_new_order=False):
        """
        update_portfolio_available performs the availability update of the concerned currency in the current portfolio
        It is called when an order is filled, created or canceled to update the "available" filled of the portfolio
        is_new_order is True when portfolio needs an update after a new order and False when portfolio needs a rollback
        after an order is cancelled
        :param order: the order to take into account
        :param is_new_order: True if this is a new order
        :return: None
        """
        if _should_update_available(order):
            self.update_portfolio_available_from_order(order, is_new_order)

    def get_portfolio_from_amount_dict(self, amount_dict):
        """
        Create a portfolio from an amount dictionary
        :param amount_dict:
        :return: the portfolio dictionary
        """
        if not all(all(isinstance(v, decimal.Decimal) for v in values.values()) for values in amount_dict.values()):
            raise RuntimeError("Portfolio has to be initialized using decimal.Decimal")
        return {
            currency: self.create_currency_asset(
                currency=currency,
                available=values[common_constants.PORTFOLIO_AVAILABLE],
                total=values[common_constants.PORTFOLIO_TOTAL]
            ).to_dict()
            for currency, values in amount_dict.items()
        }

    def _update_portfolio_data(self, currency, total_value=constants.ZERO, available_value=constants.ZERO,
                               replace_value=False):
        """
        Set new currency quantity in the portfolio
        :param currency: the currency to update
        :param total_value: the total update value
        :param available_value: the available update value
        :param replace_value: when True replace the current value instead of updating it
        :return: True if updated
        """
        try:
            if replace_value:
                return self.portfolio[currency].set(available=available_value, total=total_value)
            return self.portfolio[currency].update(available=available_value, total=total_value)
        except KeyError:
            self.portfolio[currency] = self.create_currency_asset(currency=currency,
                                                                  available=available_value, total=total_value)
            return True

    def _parse_raw_currency_asset(self, currency, raw_currency_balance):
        """
        Parse the exchange currency balance as asset
        :param currency: the currency name
        :param raw_currency_balance: the raw current currency balance
        :return: the currency asset instance
        """
        available, total = _parse_raw_currency_balance(raw_currency_balance)
        return self.create_currency_asset(currency=currency, available=available, total=total)

    def _update_raw_currency_asset(self, currency, raw_currency_balance):
        """
        Update the exchange currency asset
        :param currency: the currency name
        :param raw_currency_balance: the raw current currency balance
        :return: True if updated
        """
        available, total = _parse_raw_currency_balance(raw_currency_balance)
        return self._update_portfolio_data(currency=currency, total_value=total,
                                           available_value=available, replace_value=True)

    def reset_portfolio_available(self, reset_currency=None, reset_quantity=None):
        """
        Resets available amount with total amount
        CAREFUL: if no currency is given, resets all the portfolio !
        :param reset_currency: the currency to be reset
        :param reset_quantity: the quantity to reset
        """
        if not reset_currency:
            self._reset_all_portfolio_available()
        else:
            if reset_currency in self.portfolio:
                self._reset_currency_portfolio_available(currency_to_reset=reset_currency,
                                                         reset_quantity=reset_quantity)

    def _reset_all_portfolio_available(self):
        """
        Reset all portfolio assets available value
        """
        for currency in self.portfolio:
            self.portfolio[currency].restore_available()

    def _reset_currency_portfolio_available(self, currency_to_reset, reset_quantity):
        """
        Reset currency portfolio to available
        :param currency_to_reset: the currency to reset
        :param reset_quantity: the quantity to reset
        """
        if reset_quantity is None:
            self.portfolio[currency_to_reset].restore_available()
        else:
            self.portfolio[currency_to_reset].update(available=reset_quantity)

    def log_portfolio_update_from_order(self, order):
        """
        Log a portfolio update from an order
        :param order: the order that updated the portfolio
        """
        if order.side == enums.TradeOrderSide.BUY:
            currency_portfolio_num = order.filled_quantity - order.get_total_fees(order.currency)
            market_portfolio_num = -order.filled_quantity * order.filled_price
        else:
            currency_portfolio_num = -order.filled_quantity
            market_portfolio_num = order.filled_quantity * order.filled_price - order.get_total_fees(order.market)

        self.logger.debug(f"Portfolio updated from order | {order.currency} {currency_portfolio_num} | {order.market} "
                          f"{market_portfolio_num} | {constants.CURRENT_PORTFOLIO_STRING} {self.portfolio}")

    def log_portfolio_update_from_withdrawal(self, amount, currency):
        """
        Log a portfolio update from a withdrawal
        :param amount: withdraw quantity
        :param currency: withdraw currency
        """
        self.logger.debug(f"Portfolio updated from withdraw | {currency} -{amount}"
                          f" | {constants.CURRENT_PORTFOLIO_STRING} {self.portfolio}")


def _should_reduce_available_assets_on_fill(order):
    """
    Check if the order fill should trigger an update on the portfolio available funds.
    It's the case for orders that are not counted in available funds (stop losses). As they
    are filled, the available funds have to be reduced to keep sync with total funds.

    This is required to keep uniformity with regular order that are already reducing
    part of available funds. As we are here on the fill context of a stop-like order,
    available funds now have to be updated.
    :param order: The order to check
    :return: True if the order should update available portfolio
    """
    return not order.is_counted_in_available_funds()


def _should_update_available(order):
    """
    Check if the order has impact on availability
    :param order: The order to check
    :return: True if the order should update available portfolio
    """
    return order.is_counted_in_available_funds()


def _parse_raw_currency_balance(raw_currency_balance):
    """
    Parse the exchange balance
    Set 0 as currency value when the parsed value is None (concerning available and total values)
    :param raw_currency_balance: the current currency balance
    :return: the currency available and total as tuple
    """
    return (raw_currency_balance.get(constants.CONFIG_PORTFOLIO_FREE,
                                     raw_currency_balance.get(common_constants.PORTFOLIO_AVAILABLE,
                                                              constants.ZERO)) or constants.ZERO,
            raw_currency_balance.get(constants.CONFIG_PORTFOLIO_TOTAL,
                                     raw_currency_balance.get(common_constants.PORTFOLIO_TOTAL,
                                                              constants.ZERO)) or constants.ZERO)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library
import octobot_commons.constants as commons_constants
import octobot_commons.enums as commons_enums
import octobot_trading.enums as enums


def get_account_type_suffix_from_exchange_manager(exchange_manager) -> str:
    return get_account_type_suffix(
        exchange_manager.is_future,
        exchange_manager.is_margin,
        exchange_manager.is_option,
        exchange_manager.is_sandboxed,
        exchange_manager.is_trader_simulated
    )


def get_account_type_suffix_from_run_metadata(run_metadata) -> str:
    trading_type = run_metadata[commons_enums.DBRows.TRADING_TYPE.value]
    trader_simulator = True  # todo: update when displaying live data in strat designer
    is_sandboxed = False  # todo: update when displaying live data in strat designer
    return get_account_type_suffix(
        trading_type == enums.ExchangeTypes.FUTURE.value,
        trading_type == enums.ExchangeTypes.MARGIN.value,
        trading_type == enums.ExchangeTypes.OPTION.value,
        is_sandboxed,
        trader_simulator
    )


def get_account_type_suffix(is_future, is_margin, is_option, is_sandboxed, is_trader_simulated) -> str:
    suffix = ""
    if is_future:
        suffix = f"{suffix}_{commons_constants.CONFIG_EXCHANGE_FUTURE}"
    elif is_margin:
        suffix = f"{suffix}_{commons_constants.CONFIG_EXCHANGE_MARGIN}"
    elif is_option:
        suffix = f"{suffix}_{commons_constants.CONFIG_EXCHANGE_OPTION}"
    else:
        suffix = f"{suffix}_{commons_constants.CONFIG_EXCHANGE_SPOT}"
    if is_sandboxed:
        suffix = f"{suffix}_{commons_constants.CONFIG_EXCHANGE_SANDBOXED}"
    if is_trader_simulated:
        suffix = f"{suffix}_{commons_constants.CONFIG_SIMULATOR}"
    return suffix



# pylint: disable=E0611
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import abc
import asyncio

import octobot_commons.logging as logging
import octobot_trading.constants
import octobot_trading.enums
import octobot_trading.exchange_channel as exchange_channel
import octobot_trading.exchange_data as exchange_data


class AbstractWebsocketExchange:
    __metaclass__ = abc.ABCMeta

    REQUIRED_ACTIVATED_TENTACLES = []
    EXCHANGE_FEEDS = {}

    INIT_REQUIRING_EXCHANGE_FEEDS = set()

    # Used to ignore a feed when at least one of the corresponding feed is supported
    IGNORED_FEED_PAIRS = {}

    SUPPORTS_LIVE_PAIR_ADDITION = False

    TIME_FRAME_RELATED_FEEDS = [
        octobot_trading.enums.WebsocketFeeds.CANDLE,
        octobot_trading.enums.WebsocketFeeds.KLINE,
    ]
    # Feeds to create above which not to use websockets
    MAX_HANDLED_FEEDS = octobot_trading.constants.NO_DATA_LIMIT

    def __init__(self, config, exchange_manager):
        self.config = config

        self.exchange_manager = exchange_manager
        self.exchange = self.exchange_manager.exchange
        self.exchange_id = self.exchange_manager.id

        self.bot_mainloop = asyncio.get_event_loop()

        self.currencies = []
        self.pairs = []
        self.time_frames = []
        self.channels = []

        self.books = {}

        self.client = None
        self.name = self.get_name()
        self.logger = logging.get_logger(self._get_logger_name())

    def _get_logger_name(self):
        return f"WebSocket - {self.name}"

    def initialize(self, currencies=None, pairs=None, time_frames=None, channels=None):
        self.pairs = [self.get_exchange_pair(pair) for pair in pairs] if pairs else []
        # inner list required for cythonization
        self.channels = list(set([self.feed_to_exchange(channel) for channel in channels])) if channels else []
        self.time_frames = [
            time_frame
            for time_frame in time_frames
            if self.is_time_frame_supported(time_frame)
        ] if time_frames is not None else []
        self.currencies = currencies if currencies else []

    @classmethod
    def update_exchange_feeds(cls, exchange_manager):
        """
        Called before exchange instantiation, should be used to patch cls.EXCHANGE_FEEDS if necessary
        """

    def is_time_frame_related_feed(self, feed):
        return feed in self.TIME_FRAME_RELATED_FEEDS

    def is_time_frame_supported(self, time_frame):
        raise NotImplementedError("is_time_frame_supported is not implemented")

    def get_exchange_credentials(self):
        return self.exchange_manager.get_exchange_credentials(self.exchange_manager.exchange_name)

    async def push_to_channel(self, channel_name, *args, **kwargs):
        try:
            asyncio.run_coroutine_threadsafe(
                exchange_channel.get_chan(channel_name, self.exchange_id).get_internal_producer().push(*args, **kwargs),
                self.bot_mainloop)
        except Exception as e:
            self.logger.error(f"Push to {channel_name} failed : {e}")

    # Abstract methods
    @classmethod
    def get_name(cls):
        raise NotImplementedError("get_name not implemented")

    @classmethod
    def has_name(cls, name: str) -> bool:
        raise NotImplementedError("has_name not implemented")

    @abc.abstractmethod
    async def init_websocket(self, time_frames, trader_pairs, tentacles_setup_config):
        raise NotImplementedError("init_websocket not implemented")

    @abc.abstractmethod
    async def start_sockets(self):
        raise NotImplementedError("start_sockets not implemented")

    @abc.abstractmethod
    async def wait_sockets(self):
        raise NotImplementedError("wait_sockets not implemented")

    @abc.abstractmethod
    def update_followed_pairs(self):
        raise NotImplementedError("updated_followed_pairs not implemented")

    @abc.abstractmethod
    async def _close_and_restart_sockets(self, debounce_duration=0):
        raise NotImplementedError("_close_and_restart_sockets not implemented")

    @abc.abstractmethod
    async def stop_sockets(self):
        """
        Stops the websocket. Can be restarted
        """
        raise NotImplementedError("stop_sockets not implemented")

    @abc.abstractmethod
    async def close_sockets(self):
        """
        Closes the websocket. Can't be restarted
        """
        raise NotImplementedError("close_sockets not implemented")

    @classmethod
    def get_exchange_feed(cls, feed) -> str:
        return cls.EXCHANGE_FEEDS.get(feed, octobot_trading.enums.WebsocketFeeds.UNSUPPORTED.value)

    @classmethod
    def is_feed_requiring_init(cls, feed) -> bool:
        return feed in cls.INIT_REQUIRING_EXCHANGE_FEEDS

    @classmethod
    def is_feed_supported(cls, feed_name) -> bool:
        return feed_name != octobot_trading.enums.WebsocketFeeds.UNSUPPORTED.value

    @classmethod
    def should_ignore_feed(cls, feed):
        """
        Checks if a feed should be ignored
        :param feed: the feed (instance of octobot_trading.enums.WebsocketFeeds)
        :return: True when the feed is already covered by another one
        """
        ignored_feed_candidate_list = cls.IGNORED_FEED_PAIRS.get(feed, [])
        if not ignored_feed_candidate_list:
            return False
        for feed_candidate, feed_value in cls.EXCHANGE_FEEDS.items():
            if feed_candidate in ignored_feed_candidate_list and \
                    cls.is_feed_supported(feed_value):
                return True
        return False

    def feed_to_exchange(self, feed):
        feed_name: str = self.get_exchange_feed(feed)
        if not self.is_feed_supported(feed_name):
            self.logger.error("{} is not supported on {}".format(feed, self.get_name()))
            raise ValueError(f"{feed} is not supported on {self.get_name()}")
        return feed

    def get_book_instance(self, symbol):
        try:
            return self.books[symbol]
        except KeyError:
            self.books[symbol] = exchange_data.OrderBookManager()
            return self.books[symbol]

    def get_pair_from_exchange(self, pair):
        raise NotImplementedError("get_pair_from_exchange is not implemented")

    def get_exchange_pair(self, pair):
        raise NotImplementedError("get_exchange_pair is not implemented")

    async def reset(self):
        raise NotImplementedError("reset is not implemented")

    def add_pairs(self, pairs, watching_only=False):
        """
        Add new pairs to self.filtered_pairs
        :param pairs: the list of pair to add
        :param watching_only: if pairs are for watching or trading purpose
        """
        raise NotImplementedError("add_pairs is not implemented")

    def remove_pairs(self, pairs, watching_only=False):
        """
        Remove pairs from self.filtered_pairs
        :param pairs: the list of pair to remove
        :param watching_only: if pairs are for watching or trading purpose
        """
        raise NotImplementedError("remove_pairs is not implemented")

    def get_max_handled_pair_with_time_frame(self):
        """
        :return: the maximum number of simultaneous pairs * time_frame that this exchange can handle.
        """
        return octobot_trading.constants.INFINITE_MAX_HANDLED_PAIRS_WITH_TIMEFRAME

    @classmethod
    def get_feeds_count(cls, pairs, time_frames) -> int:
        raise NotImplementedError(f"get_feeds_count is not implemented")

    def _should_authenticate(self):
        api_key, api_secret, _, _, _ = self.get_exchange_credentials()
        return not self.exchange_manager.without_auth \
            and not self.exchange_manager.is_trader_simulated \
            and api_key and api_secret

    def clear(self):
        self.exchange = None
        self.exchange_manager = None



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import octobot_trading.enums as enums
import octobot_trading.exchange_data as exchange_data

def is_inverse_future_contract(contract_type):
    return exchange_data.FutureContract(None, None, contract_type).is_inverse_contract()


def is_perpetual_future_contract(contract_type):
    return exchange_data.FutureContract(None, None, contract_type).is_perpetual_contract()

def is_inverse_option_contract(contract_type):
    return exchange_data.OptionContract(None, None, contract_type).is_inverse_contract()


def is_perpetual_option_contract(contract_type):
    return exchange_data.OptionContract(None, None, contract_type).is_perpetual_contract()


def get_pair_contracts(exchange_manager) -> dict:
    return exchange_manager.exchange.pair_contracts


def is_handled_contract(contract) -> bool:
    return contract.is_handled_contract()


def ensure_supported_contract_configuration(exchange_manager, pair: str):
    get_pair_contracts(exchange_manager)[pair].ensure_supported_configuration()

"""
Deprecated: Use has_pair_contract instead
"""
def has_pair_future_contract(exchange_manager, pair: str) -> bool:
    return exchange_manager.exchange.has_pair_future_contract(pair)

def has_pair_contract(exchange_manager, pair: str) -> bool:
    return exchange_manager.exchange.has_pair_contract(pair)

def update_pair_contract(exchange_manager, pair: str, leverage: decimal.Decimal):
    get_pair_contracts(exchange_manager)[pair].current_leverage = leverage


def load_pair_contract(exchange_manager, contract_dict: dict):
    exchange_data.update_future_contract_from_dict(exchange_manager, contract_dict)


def create_default_future_contract(
    pair: str, leverage: decimal.Decimal, contract_type: enums.FutureContractType, position_mode: enums.PositionMode
) -> exchange_data.FutureContract:
    return exchange_data.create_default_future_contract(pair, leverage, contract_type, position_mode)

def create_default_option_contract(
    pair: str, leverage: decimal.Decimal, contract_type: enums.OptionContractType, position_mode: enums.PositionMode
) -> exchange_data.OptionContract:
    return exchange_data.create_default_option_contract(pair, leverage, contract_type, position_mode)
